import os
import unittest
from unittest import mock

from sqlglot import parse_one, transpile
from sqlglot.errors import ErrorLevel, ParseError, UnsupportedError
from sqlglot.helper import logger as helper_logger
from sqlglot.parser import logger as parser_logger
from tests.helpers import (
    assert_logger_contains,
    load_sql_fixture_pairs,
    load_sql_fixtures,
)


class TestTranspile(unittest.TestCase):
    file_dir = os.path.dirname(__file__)
    fixtures_dir = os.path.join(file_dir, "fixtures")
    maxDiff = None

    def validate(self, sql, target, **kwargs):
        self.assertEqual(transpile(sql, **kwargs)[0], target)

    def test_weird_chars(self):
        self.assertEqual(transpile("0Êß")[0], "0 AS Êß")
        self.assertEqual(
            # Ideographic space after SELECT (\u3000)
            transpile("SELECT　* FROM t WHERE c = 1")[0],
            "SELECT * FROM t WHERE c = 1",
        )

    def test_alias(self):
        self.assertEqual(transpile("SELECT SUM(y) KEEP")[0], "SELECT SUM(y) AS KEEP")
        self.assertEqual(transpile("SELECT 1 overwrite")[0], "SELECT 1 AS overwrite")
        self.assertEqual(transpile("SELECT 1 is")[0], "SELECT 1 AS is")
        self.assertEqual(transpile("SELECT 1 current_time")[0], "SELECT 1 AS current_time")
        self.assertEqual(
            transpile("SELECT 1 current_timestamp")[0], "SELECT 1 AS current_timestamp"
        )
        self.assertEqual(transpile("SELECT 1 current_date")[0], "SELECT 1 AS current_date")
        self.assertEqual(transpile("SELECT 1 current_datetime")[0], "SELECT 1 AS current_datetime")
        self.assertEqual(transpile("SELECT 1 row")[0], "SELECT 1 AS row")

        self.assertEqual(
            transpile("SELECT 1 FROM a.b.table1 t UNPIVOT((c3) FOR c4 IN (a, b))")[0],
            "SELECT 1 FROM a.b.table1 AS t UNPIVOT((c3) FOR c4 IN (a, b))",
        )

        for key in ("union", "from", "join"):
            with self.subTest(f"alias {key}"):
                self.validate(f"SELECT x AS {key}", f"SELECT x AS {key}")
                self.validate(f'SELECT x "{key}"', f'SELECT x AS "{key}"')

                with self.assertRaises(ParseError):
                    self.validate(f"SELECT x {key}", "")

    def test_unary(self):
        self.validate("+++1", "1")
        self.validate("+-1", "-1")
        self.validate("+- - -1", "- - -1")

    def test_paren(self):
        with self.assertRaises(ParseError):
            transpile("1 + (2 + 3")
            transpile("select f(")

    def test_some(self):
        self.validate(
            "SELECT * FROM x WHERE a = SOME (SELECT 1)",
            "SELECT * FROM x WHERE a = ANY(SELECT 1)",
        )

    def test_leading_comma(self):
        self.validate(
            "SELECT a, b, c FROM (SELECT a, b, c FROM t)",
            "SELECT\n"
            "    a\n"
            "    , b\n"
            "    , c\n"
            "FROM (\n"
            "    SELECT\n"
            "        a\n"
            "        , b\n"
            "        , c\n"
            "    FROM t\n"
            ")",
            leading_comma=True,
            pretty=True,
            pad=4,
            indent=4,
        )
        self.validate(
            "SELECT FOO, BAR, BAZ",
            "SELECT\n  FOO\n  , BAR\n  , BAZ",
            leading_comma=True,
            pretty=True,
        )
        self.validate(
            "SELECT FOO, /*x*/\nBAR, /*y*/\nBAZ",
            "SELECT\n  FOO /* x */\n  , BAR /* y */\n  , BAZ",
            leading_comma=True,
            pretty=True,
        )
        # without pretty, this should be a no-op
        self.validate(
            "SELECT FOO, BAR, BAZ",
            "SELECT FOO, BAR, BAZ",
            leading_comma=True,
        )

    def test_space(self):
        self.validate("SELECT MIN(3)>MIN(2)", "SELECT MIN(3) > MIN(2)")
        self.validate("SELECT MIN(3)>=MIN(2)", "SELECT MIN(3) >= MIN(2)")
        self.validate("SELECT 1>0", "SELECT 1 > 0")
        self.validate("SELECT 3>=3", "SELECT 3 >= 3")
        self.validate("SELECT a\r\nFROM b", "SELECT a FROM b")

    def test_comments(self):
        self.validate(
            "select /* asfd /* asdf */ asdf */ 1",
            "/* asfd / * asdf * / asdf */ SELECT 1",
        )
        self.validate(
            "SELECT c /* foo */ AS alias",
            "SELECT c AS alias /* foo */",
        )
        self.validate(
            "SELECT c AS /* foo */ (a, b, c) FROM t",
            "SELECT c AS (a, b, c) /* foo */ FROM t",
        )
        self.validate(
            "SELECT * FROM t1\n/*x*/\nUNION ALL SELECT * FROM t2",
            "SELECT * FROM t1 /* x */ UNION ALL SELECT * FROM t2",
        )
        self.validate(
            "/* comment */ SELECT * FROM a UNION SELECT * FROM b",
            "/* comment */ SELECT * FROM a UNION SELECT * FROM b",
        )
        self.validate(
            "SELECT * FROM t1\n/*x*/\nINTERSECT ALL SELECT * FROM t2",
            "SELECT * FROM t1 /* x */ INTERSECT ALL SELECT * FROM t2",
        )
        self.validate(
            "SELECT\n  foo\n/* comments */\n;",
            "SELECT foo /* comments */",
        )
        self.validate(
            "SELECT * FROM a INNER /* comments */ JOIN b",
            "SELECT * FROM a /* comments */ INNER JOIN b",
        )
        self.validate(
            "SELECT * FROM a LEFT /* comment 1 */ OUTER /* comment 2 */ JOIN b",
            "SELECT * FROM a /* comment 1 */ /* comment 2 */ LEFT OUTER JOIN b",
        )
        self.validate(
            "SELECT CASE /* test */ WHEN a THEN b ELSE c END",
            "SELECT CASE WHEN a THEN b ELSE c END /* test */",
        )
        self.validate("SELECT 1 /*/2 */", "SELECT 1 /* /2 */")
        self.validate("SELECT */*comment*/", "SELECT * /* comment */")
        self.validate(
            "SELECT * FROM table /*comment 1*/ /*comment 2*/",
            "SELECT * FROM table /* comment 1 */ /* comment 2 */",
        )
        self.validate("SELECT 1 FROM foo -- comment", "SELECT 1 FROM foo /* comment */")
        self.validate(
            "SELECT * FROM\n/* comment */\ndb.schema1.tbl PIVOT (SUM(a) FOR b IN ('x', 'y'))",
            "SELECT * FROM db.schema1.tbl PIVOT(SUM(a) FOR b IN ('x', 'y')) /* comment */",
        )
        self.validate("SELECT --+5\nx FROM foo", "/* +5 */ SELECT x FROM foo")
        self.validate("SELECT --!5\nx FROM foo", "/* !5 */ SELECT x FROM foo")
        self.validate(
            "SELECT 1 /* inline */ FROM foo -- comment",
            "SELECT 1 /* inline */ FROM foo /* comment */",
        )
        self.validate(
            "SELECT FUN(x) /*x*/, [1,2,3] /*y*/", "SELECT FUN(x) /* x */, ARRAY(1, 2, 3) /* y */"
        )
        self.validate(
            """
            SELECT 1 -- comment
            FROM foo -- comment
            """,
            "SELECT 1 /* comment */ FROM foo /* comment */",
        )
        self.validate(
            """
            SELECT 1 /* big comment
             like this */
            FROM foo -- comment
            """,
            """SELECT 1 /* big comment
             like this */ FROM foo /* comment */""",
        )
        self.validate(
            "select x from foo --       x",
            "SELECT x FROM foo /*       x */",
        )
        self.validate(
            """select x, --
            from foo""",
            "SELECT x FROM foo",
        )
        self.validate(
            """
-- comment 1
-- comment 2
-- comment 3
SELECT * FROM foo
            """,
            "/* comment 1 */ /* comment 2 */ /* comment 3 */ SELECT * FROM foo",
        )
        self.validate(
            """
-- comment 1
-- comment 2
-- comment 3
SELECT * FROM foo""",
            """/* comment 1 */
/* comment 2 */
/* comment 3 */
SELECT
  *
FROM foo""",
            pretty=True,
        )
        self.validate(
            """
SELECT * FROM tbl /*line1
line2
line3*/ /*another comment*/ where 1=1 -- comment at the end""",
            """SELECT * FROM tbl /* line1
line2
line3 */ /* another comment */ WHERE 1 = 1 /* comment at the end */""",
        )
        self.validate(
            """
SELECT * FROM tbl /*line1
line2
line3*/ /*another comment*/ where 1=1 -- comment at the end""",
            """SELECT
  *
FROM tbl /* line1
line2
line3 */ /* another comment */
WHERE
  1 = 1 /* comment at the end */""",
            pretty=True,
        )
        self.validate(
            """
            /* multi
               line
               comment
            */
            SELECT
              tbl.cola /* comment 1 */ + tbl.colb /* comment 2 */,
              CAST(x AS CHAR), # comment 3
              y               -- comment 4
            FROM
              bar /* comment 5 */,
              tbl #          comment 6
            """,
            """/* multi
               line
               comment
            */
SELECT
  tbl.cola /* comment 1 */ + tbl.colb /* comment 2 */,
  CAST(x AS CHAR), /* comment 3 */
  y /* comment 4 */
FROM bar /* comment 5 */, tbl /*          comment 6 */""",
            read="mysql",
            pretty=True,
        )
        self.validate(
            """
            SELECT a FROM b
            WHERE foo
            -- comment 1
            AND bar
            -- comment 2
            AND bla
            -- comment 3
            LIMIT 10
            ;
            """,
            "SELECT a FROM b WHERE foo AND /* comment 1 */ bar AND /* comment 2 */ bla LIMIT 10 /* comment 3 */",
        )
        self.validate(
            """
            SELECT a FROM b WHERE foo
            -- comment 1
            """,
            "SELECT a FROM b WHERE foo /* comment 1 */",
        )
        self.validate(
            """
            select a
            -- from
            from b
            -- where
            where foo
            -- comment 1
            and bar
            -- comment 2
            and bla
            """,
            """SELECT
  a
/* from */
FROM b
/* where */
WHERE
  foo AND /* comment 1 */ bar AND /* comment 2 */ bla""",
            pretty=True,
        )
        self.validate(
            """
            -- test
            WITH v AS (
              SELECT
                1 AS literal
            )
            SELECT
              *
            FROM v
            """,
            """/* test */
WITH v AS (
  SELECT
    1 AS literal
)
SELECT
  *
FROM v""",
            pretty=True,
        )
        self.validate(
            "(/* 1 */ 1 ) /* 2 */",
            "(1) /* 1 */ /* 2 */",
        )
        self.validate(
            "select * from t where not a in (23) /*test*/ and b in (14)",
            "SELECT * FROM t WHERE NOT a IN (23) /* test */ AND b IN (14)",
        )
        self.validate(
            "select * from t where a in (23) /*test*/ and b in (14)",
            "SELECT * FROM t WHERE a IN (23) /* test */ AND b IN (14)",
        )
        self.validate(
            "select * from t where ((condition = 1)/*test*/)",
            "SELECT * FROM t WHERE ((condition = 1) /* test */)",
        )
        self.validate(
            "SELECT 1 // hi this is a comment",
            "SELECT 1 /* hi this is a comment */",
            read="snowflake",
        )
        self.validate(
            "-- comment\nDROP TABLE IF EXISTS foo",
            "/* comment */ DROP TABLE IF EXISTS foo",
        )
        self.validate(
            """
            -- comment1
            -- comment2

            -- comment3
            DROP TABLE IF EXISTS db.tba
            """,
            """/* comment1 */
/* comment2 */
/* comment3 */
DROP TABLE IF EXISTS db.tba""",
            pretty=True,
        )
        self.validate(
            """
            -- comment4
            CREATE TABLE db.tba AS
            SELECT a, b, c
            FROM tb_01
            WHERE
            -- comment5
              a = 1 AND b = 2 --comment6
              -- and c = 1
            -- comment7
            ;
            """,
            """/* comment4 */
CREATE TABLE db.tba AS
SELECT
  a,
  b,
  c
FROM tb_01
WHERE
  a /* comment5 */ = 1 AND b = 2 /* comment6 */ /* and c = 1 */ /* comment7 */""",
            pretty=True,
        )
        self.validate(
            """
            SELECT
               -- This is testing comments
                col,
            -- 2nd testing comments
            CASE WHEN a THEN b ELSE c END as d
            FROM t
            """,
            """SELECT
  col, /* This is testing comments */
  CASE WHEN a THEN b ELSE c END AS d /* 2nd testing comments */
FROM t""",
            pretty=True,
        )
        self.validate(
            """
            SELECT * FROM a
            -- comments
            INNER JOIN b
            """,
            """SELECT
  *
FROM a
/* comments */
INNER JOIN b""",
            pretty=True,
        )
        self.validate(
            "SELECT * FROM a LEFT /* comment 1 */ OUTER /* comment 2 */ JOIN b",
            """SELECT
  *
FROM a
/* comment 1 */
/* comment 2 */
LEFT OUTER JOIN b""",
            pretty=True,
        )
        self.validate(
            "SELECT\n  a /* sqlglot.meta case_sensitive */ -- noqa\nFROM tbl",
            """SELECT
  a /* sqlglot.meta case_sensitive */ /* noqa */
FROM tbl""",
            pretty=True,
        )
        self.validate(
            """
SELECT
  'hotel1' AS hotel,
  *
FROM dw_1_dw_1_1.exactonline_1.transactionlines
/*
    UNION ALL
    SELECT
      'Thon Partner Hotel Jølster' AS hotel,
      name,
      date,
      CAST(identifier AS VARCHAR) AS identifier,
      value
    FROM d2o_889_oupjr_1348.public.accountvalues_forecast
*/
UNION ALL
SELECT
  'hotel2' AS hotel,
  *
FROM dw_1_dw_1_1.exactonline_2.transactionlines""",
            """SELECT
  'hotel1' AS hotel,
  *
FROM dw_1_dw_1_1.exactonline_1.transactionlines
/*
    UNION ALL
    SELECT
      'Thon Partner Hotel Jølster' AS hotel,
      name,
      date,
      CAST(identifier AS VARCHAR) AS identifier,
      value
    FROM d2o_889_oupjr_1348.public.accountvalues_forecast
*/
UNION ALL
SELECT
  'hotel2' AS hotel,
  *
FROM dw_1_dw_1_1.exactonline_2.transactionlines""",
            pretty=True,
        )
        self.validate(
            """/* The result of  some calculations
 */
with
    base as (
        select
            sum(sb.hep_amount) as hep_amount,
            -- I AM REMOVED
            sum(sb.hep_budget)
            /* Budget defined in sharepoint */
            as blub
            , 1 as bla
        from gold.data_budget sb
        group by all
    )
select
    *
from base
""",
            """/* The result of  some calculations
 */
WITH base AS (
  SELECT
    SUM(sb.hep_amount) AS hep_amount,
    SUM(sb.hep_budget) /* I AM REMOVED */ AS blub, /* Budget defined in sharepoint */
    1 AS bla
  FROM gold.data_budget AS sb
  GROUP BY ALL
)
SELECT
  *
FROM base""",
            pretty=True,
        )
        self.validate(
            """-- comment
SOME_FUNC(arg IGNORE NULLS)
  OVER (PARTITION BY foo ORDER BY bla) AS col""",
            "SOME_FUNC(arg IGNORE NULLS) OVER (PARTITION BY foo ORDER BY bla) AS col /* comment */",
            pretty=True,
        )
        self.validate(
            """
            SELECT *
            FROM x
            INNER JOIN y
            -- inner join z
            LEFT JOIN z using (id)
            using (id)
            """,
            """SELECT
  *
FROM x
INNER JOIN y
  /* inner join z */
  LEFT JOIN z
    USING (id)
  USING (id)""",
            pretty=True,
        )
        self.validate(
            """with x as (
  SELECT *
  /*
NOTE: LEFT JOIN because blah blah blah
  */
  FROM a
)
select * from x""",
            """WITH x AS (
  SELECT
    *
  /*
NOTE: LEFT JOIN because blah blah blah
  */
  FROM a
)
SELECT
  *
FROM x""",
            pretty=True,
        )

        self.validate(
            """SELECT X FROM  catalog.db.table WHERE Y
        --
        AND Z""",
            """SELECT X FROM catalog.db.table WHERE Y AND Z""",
        )
        self.validate(
            """with a as /* comment */ ( select * from b) select * from a""",
            """WITH a /* comment */ AS (SELECT * FROM b) SELECT * FROM a""",
        )
        self.validate(
            """
  -- comment at the top
WITH
-- comment for tbl1
tbl1 AS (SELECT 1)
-- comment for tbl2
, tbl2 AS (SELECT 2)
-- comment for tbl3
, tbl3 AS (SELECT 3)
-- comment for final select
SELECT * FROM tbl1""",
            """/* comment at the top */
WITH tbl1 /* comment for tbl1 */ AS (
  SELECT
    1
), tbl2 /* comment for tbl2 */ AS (
  SELECT
    2
), tbl3 /* comment for tbl3 */ AS (
  SELECT
    3
)
/* comment for final select */
SELECT
  *
FROM tbl1""",
            pretty=True,
        )

        self.validate(
            """
            WITH x /* a */ AS (
              SELECT 2 AS n /* b */
              FROM (/* c */ SELECT /* c2 */ a /* d */ FROM t) AS x
            )
            SELECT * FROM x /* e */
            WHERE n >= (/* f */ SELECT MAX(x) FROM t)
            ORDER BY n /* g */
            -- h
            """,
            """WITH x /* a */ AS (
  SELECT
    2 AS n /* b */
  FROM (
    /* c */
    /* c2 */
    SELECT
      a /* d */
    FROM t
  ) AS x
)
SELECT
  *
FROM x /* e */
WHERE
  n >= (
    SELECT
      MAX(x)
    FROM t
  ) /* f */
ORDER BY
  n /* g */ /* h */""",
            pretty=True,
        )

        # Round-trip stress: multiple trailing comments on the same expression must
        # stay space-separated (same line) in pretty mode. Re-parsing would otherwise
        # detach the later ones onto the next token.
        self.validate(
            "SELECT a /* foo */ /* bar */ FROM tbl",
            """SELECT
  a /* foo */ /* bar */
FROM tbl""",
            pretty=True,
        )
        self.validate(
            "SELECT x FROM t WHERE x = 1 /* a */ /* b */ /* c */",
            """SELECT
  x
FROM t
WHERE
  x = 1 /* a */ /* b */ /* c */""",
            pretty=True,
        )

        self.validate(
            """SELECT
  *
FROM x
WHERE
  a = 1 AND /*
  hello
  world
*/ 1 = 0""",
            """SELECT
  *
FROM x
WHERE
  a = 1 AND /*
  hello
  world
*/ 1 = 0""",
            pretty=True,
        )
        self.validate(
            """SELECT
  *
FROM x
WHERE
  a = 1
  AND /*
  line1

  line3
*/ b = 2""",
            """SELECT
  *
FROM x
WHERE
  a = 1
  AND /*
  line1

  line3
*/ b = 2""",
            pretty=True,
        )

    def test_comment_single_line_with_block_close(self):
        # Single-line comments containing */ must be escaped when converted to block comments,
        # otherwise the */ prematurely closes the block comment and turns comment text into SQL.
        self.validate(
            "-- aa */ SELECT * FROM secret_table --\nSELECT 1",
            "/* aa * / SELECT * FROM secret_table -- */ SELECT 1",
        )
        self.validate(
            "-- comment */ DROP TABLE users --\nSELECT 1",
            "/* comment * / DROP TABLE users -- */ SELECT 1",
        )
        # Nested block comments have their inner markers escaped to prevent misparse on re-emit
        self.validate(
            "SELECT c /* c1 /* c2 */ c3 */",
            "SELECT c /* c1 / * c2 * / c3 */",
        )
        self.validate(
            "SELECT c /* c1 /* c2 /* c3 */ */ */",
            "SELECT c /* c1 / * c2 / * c3 * / * / */",
        )

    def test_types(self):
        self.validate("INT 1", "CAST(1 AS INT)")
        self.validate("VARCHAR 'x' y", "CAST('x' AS VARCHAR) AS y")
        self.validate("STRING 'x' y", "CAST('x' AS TEXT) AS y")
        self.validate("x::INT", "CAST(x AS INT)")
        self.validate("x::INTEGER", "CAST(x AS INT)")
        self.validate("x::INT y", "CAST(x AS INT) AS y")
        self.validate("x::INT AS y", "CAST(x AS INT) AS y")
        self.validate("x::INT::BOOLEAN", "CAST(CAST(x AS INT) AS BOOLEAN)")
        self.validate("interval::int", "CAST(interval AS INT)")
        self.validate("x::user_defined_type", "CAST(x AS user_defined_type)")
        self.validate("CAST(x::INT AS BOOLEAN)", "CAST(CAST(x AS INT) AS BOOLEAN)")
        self.validate("CAST(x AS INT)::BOOLEAN", "CAST(CAST(x AS INT) AS BOOLEAN)")

        with self.assertRaises(ParseError):
            transpile("x::z", read="clickhouse")

    def test_not_range(self):
        self.validate("a NOT LIKE b", "a NOT LIKE b")
        self.validate("a NOT BETWEEN b AND c", "NOT a BETWEEN b AND c")
        self.validate("a NOT IN (1, 2)", "NOT a IN (1, 2)")
        self.validate("a IS NOT NULL", "NOT a IS NULL")
        self.validate("a LIKE TEXT 'y'", "a LIKE CAST('y' AS TEXT)")

    def test_extract(self):
        self.validate(
            "EXTRACT(day FROM '2020-01-01'::TIMESTAMP)",
            "EXTRACT(DAY FROM CAST('2020-01-01' AS TIMESTAMP))",
        )
        self.validate(
            "EXTRACT(timezone FROM '2020-01-01'::TIMESTAMP)",
            "EXTRACT(TIMEZONE FROM CAST('2020-01-01' AS TIMESTAMP))",
        )
        self.validate(
            "EXTRACT(year FROM '2020-01-01'::TIMESTAMP WITH TIME ZONE)",
            "EXTRACT(YEAR FROM CAST('2020-01-01' AS TIMESTAMPTZ))",
        )
        self.validate(
            "extract(month from '2021-01-31'::timestamp without time zone)",
            "EXTRACT(MONTH FROM CAST('2021-01-31' AS TIMESTAMP))",
        )
        self.validate("extract(week from current_date + 2)", "EXTRACT(WEEK FROM CURRENT_DATE + 2)")
        self.validate(
            "EXTRACT(minute FROM datetime1 - datetime2)",
            "EXTRACT(MINUTE FROM datetime1 - datetime2)",
        )

    def test_if(self):
        self.validate(
            "SELECT IF(a > 1, 1, 0) FROM foo",
            "SELECT CASE WHEN a > 1 THEN 1 ELSE 0 END FROM foo",
        )
        self.validate(
            "SELECT IF a > 1 THEN b END",
            "SELECT CASE WHEN a > 1 THEN b END",
        )
        self.validate(
            "SELECT IF a > 1 THEN b ELSE c END",
            "SELECT CASE WHEN a > 1 THEN b ELSE c END",
        )
        self.validate("SELECT IF(a > 1, 1) FROM foo", "SELECT CASE WHEN a > 1 THEN 1 END FROM foo")

    def test_with(self):
        self.validate(
            "WITH a AS (SELECT 1) WITH b AS (SELECT 2) SELECT *",
            "WITH a AS (SELECT 1), b AS (SELECT 2) SELECT *",
        )
        self.validate(
            "WITH a AS (SELECT 1), WITH b AS (SELECT 2) SELECT *",
            "WITH a AS (SELECT 1), b AS (SELECT 2) SELECT *",
        )
        self.validate(
            "WITH A(filter) AS (VALUES 1, 2, 3) SELECT * FROM A WHERE filter >= 2",
            "WITH A(filter) AS (SELECT * FROM (VALUES (1), (2), (3)) AS _values) SELECT * FROM A WHERE filter >= 2",
            read="presto",
        )
        self.validate(
            "SELECT 1 UNION ALL VALUES (2)",
            "SELECT 1 UNION ALL SELECT * FROM (VALUES (2)) AS _values",
        )
        self.validate(
            "VALUES (1) AS v(a) UNION ALL SELECT 2",
            "SELECT * FROM (VALUES (1)) AS v(a) UNION ALL SELECT 2",
        )
        self.validate(
            "VALUES (1) UNION SELECT * FROM x",
            "SELECT * FROM (VALUES (1)) AS _values UNION SELECT * FROM x",
        )
        self.validate(
            "WITH RECURSIVE T(n) AS (VALUES (1) UNION ALL SELECT n + 1 FROM t WHERE n < 100) SELECT SUM(n) FROM t",
            "WITH RECURSIVE T(n) AS (SELECT * FROM (VALUES (1)) AS _values UNION ALL SELECT n + 1 FROM t WHERE n < 100) SELECT SUM(n) FROM t",
        )
        self.validate(
            "SELECT BOOL_OR(a > 10) FROM (VALUES 1, 2, 15) AS T(a)",
            "SELECT BOOL_OR(a > 10) FROM (VALUES (1), (2), (15)) AS T(a)",
            read="presto",
        )

    def test_alter(self):
        self.validate(
            "ALTER TABLE integers ADD k INTEGER",
            "ALTER TABLE integers ADD COLUMN k INT",
        )
        self.validate(
            "ALTER TABLE integers ALTER i TYPE VARCHAR",
            "ALTER TABLE integers ALTER COLUMN i SET DATA TYPE VARCHAR",
        )
        self.validate(
            "ALTER TABLE integers ALTER i TYPE VARCHAR COLLATE foo USING bar",
            "ALTER TABLE integers ALTER COLUMN i SET DATA TYPE VARCHAR COLLATE foo USING bar",
        )

    def test_time(self):
        self.validate("INTERVAL '1 day'", "INTERVAL '1' DAY")
        self.validate("INTERVAL '1 days' * 5", "INTERVAL '1' DAYS * 5")
        self.validate("5 * INTERVAL '1 day'", "5 * INTERVAL '1' DAY")
        self.validate("INTERVAL 1 day", "INTERVAL '1' DAY")
        self.validate("INTERVAL 2 months", "INTERVAL '2' MONTHS")
        self.validate("TIMESTAMP '2020-01-01'", "CAST('2020-01-01' AS TIMESTAMP)")
        self.validate("TIMESTAMP WITH TIME ZONE '2020-01-01'", "CAST('2020-01-01' AS TIMESTAMPTZ)")
        self.validate(
            "TIMESTAMP(9) WITH TIME ZONE '2020-01-01'",
            "CAST('2020-01-01' AS TIMESTAMPTZ(9))",
        )
        self.validate(
            "TIMESTAMP WITHOUT TIME ZONE '2020-01-01'",
            "CAST('2020-01-01' AS TIMESTAMP)",
        )
        self.validate("'2020-01-01'::TIMESTAMP", "CAST('2020-01-01' AS TIMESTAMP)")
        self.validate(
            "'2020-01-01'::TIMESTAMP WITHOUT TIME ZONE",
            "CAST('2020-01-01' AS TIMESTAMP)",
        )
        self.validate(
            "'2020-01-01'::TIMESTAMP WITH TIME ZONE",
            "CAST('2020-01-01' AS TIMESTAMPTZ)",
        )
        self.validate(
            "timestamp with time zone '2025-11-20 00:00:00+00' AT TIME ZONE 'Africa/Cairo'",
            "CAST('2025-11-20 00:00:00+00' AS TIMESTAMPTZ) AT TIME ZONE 'Africa/Cairo'",
        )

        self.validate("DATE '2020-01-01'", "CAST('2020-01-01' AS DATE)")
        self.validate("'2020-01-01'::DATE", "CAST('2020-01-01' AS DATE)")
        self.validate("STR_TO_TIME('x', 'y')", "STRPTIME('x', 'y')", write="duckdb")
        self.validate("STR_TO_UNIX('x', 'y')", "EPOCH(STRPTIME('x', 'y'))", write="duckdb")
        self.validate("TIME_TO_STR(x, 'y')", "STRFTIME(x, 'y')", write="duckdb")
        self.validate("TIME_TO_UNIX(x)", "EPOCH(x)", write="duckdb")
        self.validate(
            "UNIX_TO_STR(123, 'y')",
            "STRFTIME(TO_TIMESTAMP(123), 'y')",
            write="duckdb",
        )
        self.validate(
            "UNIX_TO_TIME(123)",
            "TO_TIMESTAMP(123)",
            write="duckdb",
        )

        self.validate(
            "STR_TO_TIME(x, 'y')",
            "CAST(FROM_UNIXTIME(UNIX_TIMESTAMP(x, 'y')) AS TIMESTAMP)",
            write="hive",
        )
        self.validate(
            "STR_TO_TIME(x, 'yyyy-MM-dd HH:mm:ss')",
            "CAST(x AS TIMESTAMP)",
            write="hive",
        )
        self.validate(
            "STR_TO_TIME(x, 'yyyy-MM-dd')",
            "CAST(x AS TIMESTAMP)",
            write="hive",
        )

        self.validate(
            "STR_TO_UNIX('x', 'y')",
            "UNIX_TIMESTAMP('x', 'y')",
            write="hive",
        )
        self.validate("TIME_TO_STR(x, 'y')", "DATE_FORMAT(x, 'y')", write="hive")

        self.validate("TIME_STR_TO_TIME(x)", "TIME_STR_TO_TIME(x)", write=None)
        self.validate(
            "TIME_STR_TO_TIME(x, 'America/Los_Angeles')",
            "TIME_STR_TO_TIME(x, 'America/Los_Angeles')",
            write=None,
        )
        self.validate("TIME_STR_TO_UNIX(x)", "TIME_STR_TO_UNIX(x)", write=None)
        self.validate("TIME_TO_TIME_STR(x)", "CAST(x AS TEXT)", write=None)
        self.validate("TIME_TO_STR(x, 'y')", "TIME_TO_STR(x, 'y')", write=None)
        self.validate("TIME_TO_UNIX(x)", "TIME_TO_UNIX(x)", write=None)
        self.validate("UNIX_TO_STR(x, 'y')", "UNIX_TO_STR(x, 'y')", write=None)
        self.validate("UNIX_TO_TIME(x)", "UNIX_TO_TIME(x)", write=None)
        self.validate("UNIX_TO_TIME_STR(x)", "UNIX_TO_TIME_STR(x)", write=None)
        self.validate("TIME_STR_TO_DATE(x)", "TIME_STR_TO_DATE(x)", write=None)

        self.validate("TIME_STR_TO_DATE(x)", "TO_DATE(x)", write="hive")
        self.validate("UNIX_TO_STR(x, 'yyyy-MM-dd HH:mm:ss')", "FROM_UNIXTIME(x)", write="hive")
        self.validate("STR_TO_UNIX(x, 'yyyy-MM-dd HH:mm:ss')", "UNIX_TIMESTAMP(x)", write="hive")
        self.validate("IF(x > 1, x + 1)", "IF(x > 1, x + 1)", write="presto")
        self.validate("IF(x > 1, 1 + 1)", "IF(x > 1, 1 + 1)", write="hive")
        self.validate("IF(x > 1, 1, 0)", "IF(x > 1, 1, 0)", write="hive")

        self.validate(
            "TIME_TO_UNIX(x)",
            "UNIX_TIMESTAMP(x)",
            write="hive",
        )
        self.validate("UNIX_TO_STR(123, 'y')", "FROM_UNIXTIME(123, 'y')", write="hive")
        self.validate(
            "UNIX_TO_TIME(123)",
            "FROM_UNIXTIME(123)",
            write="hive",
        )

        self.validate("STR_TO_TIME('x', 'y')", "DATE_PARSE('x', 'y')", write="presto")
        self.validate(
            "STR_TO_UNIX('x', 'y')",
            "TO_UNIXTIME(COALESCE(TRY(DATE_PARSE(CAST('x' AS VARCHAR), 'y')), PARSE_DATETIME(DATE_FORMAT(CAST('x' AS TIMESTAMP), 'y'), 'y')))",
            write="presto",
        )
        self.validate("TIME_TO_STR(x, 'y')", "DATE_FORMAT(x, 'y')", write="presto")
        self.validate("TIME_TO_UNIX(x)", "TO_UNIXTIME(x)", write="presto")
        self.validate(
            "UNIX_TO_STR(123, 'y')",
            "DATE_FORMAT(FROM_UNIXTIME(123), 'y')",
            write="presto",
        )
        self.validate("UNIX_TO_TIME(123)", "FROM_UNIXTIME(123)", write="presto")

        self.validate("STR_TO_TIME('x', 'y')", "TO_TIMESTAMP('x', 'y')", write="spark")
        self.validate("STR_TO_UNIX('x', 'y')", "UNIX_TIMESTAMP('x', 'y')", write="spark")
        self.validate("TIME_TO_STR(x, 'y')", "DATE_FORMAT(x, 'y')", write="spark")

        self.validate(
            "TIME_TO_UNIX(x)",
            "UNIX_TIMESTAMP(x)",
            write="spark",
        )
        self.validate("UNIX_TO_STR(123, 'y')", "FROM_UNIXTIME(123, 'y')", write="spark")
        self.validate(
            "UNIX_TO_TIME(123)",
            "CAST(FROM_UNIXTIME(123) AS TIMESTAMP)",
            write="spark",
        )
        self.validate(
            "CREATE TEMPORARY TABLE test AS SELECT 1",
            "CREATE TEMPORARY VIEW test AS SELECT 1",
            write="spark2",
        )

    def test_index_offset(self):
        with self.assertLogs(helper_logger) as cm:
            self.validate("x[0]", "x[1]", write="presto", identity=False)
            self.validate("x[1]", "x[0]", read="presto", identity=False)

            self.validate("x[x - 1]", "x[x - 1]", write="presto", identity=False)
            self.validate(
                "x[array_size(y) - 1]",
                "x[(CARDINALITY(y) - 1) + 1]",
                write="presto",
                identity=False,
            )
            self.validate("x[3 - 1]", "x[3]", write="presto", identity=False)
            self.validate("MAP(a, b)[0]", "MAP(a, b)[0]", write="presto", identity=False)

            self.assertEqual(
                cm.output,
                [
                    "INFO:sqlglot:Applying array index offset (1)",
                    "INFO:sqlglot:Applying array index offset (-1)",
                    "INFO:sqlglot:Applying array index offset (1)",
                    "INFO:sqlglot:Applying array index offset (1)",
                ],
            )

    def test_identify_lambda(self):
        self.validate("x(y -> y)", 'X("y" -> "y")', identify=True)

    def test_identity(self):
        self.assertEqual(transpile("")[0], "")
        for sql in load_sql_fixtures("identity.sql"):
            with self.subTest(sql):
                self.assertEqual(transpile(sql)[0], sql.strip())

    def test_command_identity(self):
        for sql in (
            "ALTER AGGREGATE bla(foo) OWNER TO CURRENT_USER",
            "ALTER DOMAIN foo VALIDATE CONSTRAINT bla",
            "ALTER ROLE CURRENT_USER WITH REPLICATION",
            "ALTER RULE foo ON bla RENAME TO baz",
            "ALTER SEQUENCE IF EXISTS baz RESTART WITH boo",
            "ALTER TABLE integers DROP PRIMARY KEY",
            "ALTER TABLE table1 MODIFY COLUMN name1 SET TAG foo='bar'",
            "ALTER TABLE table1 RENAME COLUMN c1 AS c2",
            "ALTER TABLE table1 RENAME COLUMN c1 TO c2, c2 TO c3",
            "ALTER TABLE table1 RENAME COLUMN c1 c2",
            "ALTER TYPE electronic_mail RENAME TO email",
            "ALTER schema doo",
            "CALL catalog.system.iceberg_procedure_name(named_arg_1 => 'arg_1', named_arg_2 => 'arg_2')",
            "COMMENT ON ACCESS METHOD gin IS 'GIN index access method'",
            "CREATE OR REPLACE STAGE",
            "EXECUTE statement",
            "EXPLAIN SELECT * FROM x",
            "LOAD foo",
            "OPTIMIZE TABLE y",
            "PREPARE statement",
            "SET -v",
            "SET @user OFF",
            "SHOW TABLES",
            "VACUUM FREEZE my_table",
        ):
            with self.subTest(sql):
                with self.assertLogs(parser_logger) as cm:
                    self.assertEqual(transpile(sql)[0], sql)
                    assert f"'{sql[:100]}' contains unsupported syntax" in cm.output[0]

    def test_normalize_name(self):
        self.assertEqual(
            transpile("cardinality(x)", read="presto", write="presto", normalize_functions="lower")[
                0
            ],
            "cardinality(x)",
        )

    def test_partial(self):
        for sql in load_sql_fixtures("partial.sql"):
            with self.subTest(sql):
                self.assertEqual(transpile(sql, error_level=ErrorLevel.IGNORE)[0], sql.strip())

    def test_pretty(self):
        for _, sql, pretty in load_sql_fixture_pairs("pretty.sql"):
            with self.subTest(sql[:100]):
                generated = transpile(sql, pretty=True)[0]
                self.assertEqual(generated, pretty)
                self.assertEqual(parse_one(sql), parse_one(pretty))

    def test_pretty_line_breaks(self):
        self.assertEqual(transpile("SELECT '1\n2'", pretty=True)[0], "SELECT\n  '1\n2'")
        self.assertEqual(
            transpile("SELECT '1\n2'", pretty=True, unsupported_level=ErrorLevel.IGNORE)[0],
            "SELECT\n  '1\n2'",
        )
        self.assertEqual(transpile('SELECT "1\n2"', pretty=True)[0], 'SELECT\n  "1\n2"')
        self.assertEqual(
            transpile('SELECT "Product\n(Foo, Bar)" AS x FROM t', pretty=True)[0],
            'SELECT\n  "Product\n(Foo, Bar)" AS x\nFROM t',
        )

    def test_sql_security(self):
        sqlglot_sql = "CREATE VIEW v SQL SECURITY INVOKER AS SELECT 1"

        for dialect, sql in [
            ("clickhouse", "CREATE VIEW v SQL SECURITY INVOKER AS SELECT 1"),
            ("trino", "CREATE VIEW v SECURITY INVOKER AS SELECT 1"),
            ("presto", "CREATE VIEW v SECURITY INVOKER AS SELECT 1"),
            ("starrocks", "CREATE VIEW v SECURITY INVOKER AS SELECT 1"),
            ("mysql", "CREATE SQL SECURITY INVOKER VIEW v AS SELECT 1"),
        ]:
            with self.subTest(dialect):
                self.validate(sql, read=dialect, identity=False, target=sqlglot_sql)
                self.validate(sql, write=dialect, identity=False, target=sql)

    @mock.patch("sqlglot.parser.logger")
    def test_error_level(self, logger):
        invalid = "x + 1. ("
        expected_messages = [
            "Required keyword: 'expressions' missing for <class 'sqlglot.expressions.core.Aliases'>. Line 1, Col: 8.\n  x + 1. \033[4m(\033[0m",
            "Expecting ). Line 1, Col: 8.\n  x + 1. \033[4m(\033[0m",
        ]
        expected_errors = [
            {
                "description": "Required keyword: 'expressions' missing for <class 'sqlglot.expressions.core.Aliases'>",
                "line": 1,
                "col": 8,
                "start_context": "x + 1. ",
                "highlight": "(",
                "end_context": "",
                "into_expression": None,
            },
            {
                "description": "Expecting )",
                "line": 1,
                "col": 8,
                "start_context": "x + 1. ",
                "highlight": "(",
                "end_context": "",
                "into_expression": None,
            },
        ]

        transpile(invalid, error_level=ErrorLevel.WARN)
        for error in expected_messages:
            assert_logger_contains(error, logger)

        with self.assertRaises(ParseError) as ctx:
            transpile(invalid, error_level=ErrorLevel.IMMEDIATE)

        self.assertEqual(str(ctx.exception), expected_messages[0])
        self.assertEqual(ctx.exception.errors[0], expected_errors[0])

        with self.assertRaises(ParseError) as ctx:
            transpile(invalid, error_level=ErrorLevel.RAISE)

        self.assertEqual(str(ctx.exception), "\n\n".join(expected_messages))
        self.assertEqual(ctx.exception.errors, expected_errors)

        more_than_max_errors = "(((("
        expected_messages = (
            "Required keyword: 'this' missing for <class 'sqlglot.expressions.core.Paren'>. Line 1, Col: 4.\n  (((\033[4m(\033[0m\n\n"
            "Expecting ). Line 1, Col: 4.\n  (((\033[4m(\033[0m\n\n"
            "Expecting ). Line 1, Col: 4.\n  (((\033[4m(\033[0m\n\n"
            "... and 2 more"
        )
        expected_errors = [
            {
                "description": "Required keyword: 'this' missing for <class 'sqlglot.expressions.core.Paren'>",
                "line": 1,
                "col": 4,
                "start_context": "(((",
                "highlight": "(",
                "end_context": "",
                "into_expression": None,
            },
            {
                "description": "Expecting )",
                "line": 1,
                "col": 4,
                "start_context": "(((",
                "highlight": "(",
                "end_context": "",
                "into_expression": None,
            },
        ]
        # Also expect three trailing structured errors that match the first
        expected_errors += [expected_errors[1]] * 3

        with self.assertRaises(ParseError) as ctx:
            transpile(more_than_max_errors, error_level=ErrorLevel.RAISE)

        self.assertEqual(str(ctx.exception), expected_messages)
        self.assertEqual(ctx.exception.errors, expected_errors)

    @mock.patch("sqlglot.generator.logger")
    def test_unsupported_level(self, logger):
        def unsupported(level):
            transpile(
                "SELECT MAP(a, b), MAP(a, b), MAP(a, b), MAP(a, b)",
                read="presto",
                write="hive",
                unsupported_level=level,
            )

        error = "Cannot convert array columns into map."

        unsupported(ErrorLevel.WARN)
        assert_logger_contains("\n".join([error] * 4), logger, level="warning")

        with self.assertRaises(UnsupportedError) as ctx:
            unsupported(ErrorLevel.RAISE)
        self.assertEqual(str(ctx.exception).count(error), 3)

        with self.assertRaises(UnsupportedError) as ctx:
            unsupported(ErrorLevel.IMMEDIATE)
        self.assertEqual(str(ctx.exception).count(error), 1)

    def test_recursion(self):
        sql = "1 AND 2 OR 3 AND " * 1000
        sql += "4"
        self.assertEqual(len(parse_one(sql).sql()), 17001)



import unittest

from sqlglot.anonymize import anonymize, render
from sqlglot.tokens import Tokenizer, TokenType


class TestAnonymize(unittest.TestCase):
    def assert_anonymized(self, sql, expected, dialect=None):
        """expected: list of (TokenType, text[, comments]) for the full anonymized token stream."""
        tokens = anonymize(sql, dialect)
        self.assertEqual(
            [(t.token_type, t.text, t.comments) for t in tokens],
            [entry + ([],) if len(entry) == 2 else entry for entry in expected],
        )

    def test_identifiers_rewritten(self):
        self.assert_anonymized(
            "SELECT foo FROM bar",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "aaa"),  # foo
                (TokenType.FROM, "FROM"),
                (TokenType.VAR, "aab"),  # bar
            ],
        )

    def test_consistent_within_call(self):
        self.assert_anonymized(
            "SELECT a, a, b FROM t",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "a"),  # a
                (TokenType.COMMA, ","),
                (TokenType.VAR, "a"),  # a, same alias as above
                (TokenType.COMMA, ","),
                (TokenType.VAR, "b"),  # b
                (TokenType.FROM, "FROM"),
                (TokenType.VAR, "c"),  # t
            ],
        )

    def test_quoted_identifier_shares_alias(self):
        self.assert_anonymized(
            'SELECT foo, "foo"',
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "aaa"),
                (TokenType.COMMA, ","),
                (TokenType.IDENTIFIER, "aaa"),
            ],
        )

    def test_strings_rewritten(self):
        self.assert_anonymized(
            "SELECT 'hello', 'world' FROM t",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.STRING, "aaaaa"),  # hello
                (TokenType.COMMA, ","),
                (TokenType.STRING, "aaaab"),  # world
                (TokenType.FROM, "FROM"),
                (TokenType.VAR, "c"),  # t
            ],
        )

    def test_whitespace(self):
        self.assert_anonymized(
            "SELECT 'line1\nline2', 'spam  eggs', 'a\tb', \"my table\" FROM t",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.STRING, "aaaaa\naaaaa"),  # line1 + newline + line2 (line break kept)
                (TokenType.COMMA, ","),
                (TokenType.STRING, "aaaa  aaab"),  # spam  eggs (spaces kept)
                (TokenType.COMMA, ","),
                (TokenType.STRING, "a\tc"),  # a\tb (tab kept)
                (TokenType.COMMA, ","),
                (TokenType.IDENTIFIER, "aa aaaad"),  # "my table" (space kept)
                (TokenType.FROM, "FROM"),
                (TokenType.VAR, "e"),  # t
            ],
        )

    def test_reserved_keywords_not_rewritten(self):
        # keyword tokens (e.g. window, out) aren't sensitive; they stay verbatim so
        # user SQL that trips the parser remains debuggable — while real identifiers get rewritten
        self.assert_anonymized(
            "SELECT * FROM window, OUT, apple",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.STAR, "*"),
                (TokenType.FROM, "FROM"),
                (TokenType.WINDOW, "window"),
                (TokenType.COMMA, ","),
                (TokenType.OUT, "OUT"),
                (TokenType.COMMA, ","),
                (TokenType.VAR, "aaaaa"),  # apple
            ],
        )

    def test_tokenize_error_tail_blanked(self):
        tokens = anonymize("SELECT foo, 'secret tail")
        self.assertEqual(
            [(t.token_type, t.text) for t in tokens],
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "aaa"),  # foo
                (TokenType.COMMA, ","),
                (TokenType.UNKNOWN, "'s.........."),  # 'secret tail, delimiter kept
            ],
        )

    def test_comments_redacted(self):
        self.assert_anonymized(
            "SELECT a -- secret comment\nFROM t",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "a", [" ...... ......."]),
                (TokenType.FROM, "FROM"),
                (TokenType.VAR, "b"),
            ],
        )

    def test_tpcds_query_reserialized(self):
        sql = """WITH inv
     AS (SELECT w_warehouse_name,
                w_warehouse_sk,
                i_item_sk,
                d_moy,
                stdev,
                mean,
                CASE mean
                  WHEN 0 THEN NULL
                  ELSE stdev / mean
                END cov
         FROM  (SELECT w_warehouse_name,
                       w_warehouse_sk,
                       i_item_sk,
                       d_moy,
                       Stddev_samp(inv_quantity_on_hand) stdev,
                       Avg(inv_quantity_on_hand)         mean
                FROM   inventory,
                       item,
                       warehouse,
                       date_dim
                WHERE  inv_item_sk = i_item_sk
                       AND inv_warehouse_sk = w_warehouse_sk
                       AND inv_date_sk = d_date_sk
                       AND d_year = 2002
                GROUP  BY w_warehouse_name,
                          w_warehouse_sk,
                          i_item_sk,
                          d_moy) foo
         WHERE  CASE mean
                  WHEN 0 THEN 0
                  ELSE stdev / mean
                END > 1)
SELECT inv1.w_warehouse_sk,
       inv1.i_item_sk,
       inv1.d_moy,
       inv1.mean,
       inv1.cov,
       inv2.w_warehouse_sk,
       inv2.i_item_sk,
       inv2.d_moy,
       inv2.mean,
       inv2.cov
FROM   inv inv1,
       inv inv2
WHERE  inv1.i_item_sk = inv2.i_item_sk
       AND inv1.w_warehouse_sk = inv2.w_warehouse_sk
       AND inv1.d_moy = 1
       AND inv2.d_moy = 1 + 1
ORDER  BY inv1.w_warehouse_sk,
          inv1.i_item_sk,
          inv1.d_moy,
          inv1.mean,
          inv1.cov,
          inv2.d_moy,
          inv2.mean,
          inv2.cov;"""
        expected = """WITH aaa
     AS (SELECT aaaaaaaaaaaaaaab,
                aaaaaaaaaaaaac,
                aaaaaaaad,
                aaaae,
                aaaaf,
                aaag,
                CASE aaag
                  WHEN 8 THEN NULL
                  ELSE aaaaf / aaag
                END aai
         FROM  (SELECT aaaaaaaaaaaaaaab,
                       aaaaaaaaaaaaac,
                       aaaaaaaad,
                       aaaae,
                       Stddev_samp(aaaaaaaaaaaaaaaaaaaj) aaaaf,
                       Avg(aaaaaaaaaaaaaaaaaaaj)         aaag
                FROM   aaaaaaaak,
                       aaal,
                       aaaaaaaam,
                       aaaaaaan
                WHERE  aaaaaaaaaao = aaaaaaaad
                       AND aaaaaaaaaaaaaaap = aaaaaaaaaaaaac
                       AND aaaaaaaaaaq = aaaaaaaar
                       AND aaaaas = 1019
                GROUP  BY aaaaaaaaaaaaaaab,
                          aaaaaaaaaaaaac,
                          aaaaaaaad,
                          aaaae) aau
         WHERE  CASE aaag
                  WHEN 8 THEN 8
                  ELSE aaaaf / aaag
                END > 4)
SELECT aaaw.aaaaaaaaaaaaac,
       aaaw.aaaaaaaad,
       aaaw.aaaae,
       aaaw.aaag,
       aaaw.aai,
       aaax.aaaaaaaaaaaaac,
       aaax.aaaaaaaad,
       aaax.aaaae,
       aaax.aaag,
       aaax.aai
FROM   aaa aaaw,
       aaa aaax
WHERE  aaaw.aaaaaaaad = aaax.aaaaaaaad
       AND aaaw.aaaaaaaaaaaaac = aaax.aaaaaaaaaaaaac
       AND aaaw.aaaae = 4
       AND aaax.aaaae = 4 + 4
ORDER  BY aaaw.aaaaaaaaaaaaac,
          aaaw.aaaaaaaad,
          aaaw.aaaae,
          aaaw.aaag,
          aaaw.aai,
          aaax.aaaae,
          aaax.aaag,
          aaax.aai;"""
        self.assertEqual(render(sql, anonymize(sql)), expected)
        self.assertEqual(len(expected), len(sql))

    def test_functions_anonymized(self):
        self.assert_anonymized(
            "SELECT SUM(x), my_udf(...), CURRENT_DATE, CASE WHEN 1 THEN 2 ELSE 3 END",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "SUM"),  # known function — kept
                (TokenType.L_PAREN, "("),
                (TokenType.VAR, "a"),  # x
                (TokenType.R_PAREN, ")"),
                (TokenType.COMMA, ","),
                (TokenType.VAR, "aaaaab"),  # my_udf (unknown) — masked
                (TokenType.L_PAREN, "("),
                (TokenType.DOT, "."),
                (TokenType.DOT, "."),
                (TokenType.DOT, "."),
                (TokenType.R_PAREN, ")"),
                (TokenType.COMMA, ","),
                (TokenType.CURRENT_DATE, "CURRENT_DATE"),  # no-paren function — kept
                (TokenType.COMMA, ","),
                (TokenType.CASE, "CASE"),  # keyword — kept
                (TokenType.WHEN, "WHEN"),
                (TokenType.NUMBER, "3"),  # 1
                (TokenType.THEN, "THEN"),
                (TokenType.NUMBER, "4"),  # 2
                (TokenType.ELSE, "ELSE"),
                (TokenType.NUMBER, "5"),  # 3
                (TokenType.END, "END"),
            ],
        )

        # JSON_OBJECT is registered in FUNCTION_PARSERS rather than FUNCTIONS
        self.assert_anonymized(
            "SELECT JSON_OBJECT('k', v)",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "JSON_OBJECT"),  # known function — kept
                (TokenType.L_PAREN, "("),
                (TokenType.STRING, "a"),  # 'k'
                (TokenType.COMMA, ","),
                (TokenType.VAR, "b"),  # v
                (TokenType.R_PAREN, ")"),
            ],
        )

    def test_string_family_forms_anonymized(self):
        self.assert_anonymized(
            "SELECT N'nat', $$her$$, x'4141'",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.NATIONAL_STRING, "aaa"),
                (TokenType.COMMA, ","),
                (TokenType.RAW_STRING, "aab"),
                (TokenType.COMMA, ","),
                (TokenType.HEX_STRING, "aaac"),
            ],
            "snowflake",
        )

    def test_empty_input(self):
        self.assertEqual(anonymize(""), [])
        self.assertEqual(anonymize("   "), [])

    def test_identifiers_with_spaces_stay_distinct(self):
        self.assert_anonymized(
            'SELECT "a b", "a c"',
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.IDENTIFIER, "a a"),
                (TokenType.COMMA, ","),
                (TokenType.IDENTIFIER, "a b"),
            ],
        )

    def test_number_and_string_same_text_stay_distinct(self):
        self.assert_anonymized(
            "SELECT 123, '123'",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.NUMBER, "100"),
                (TokenType.COMMA, ","),
                (TokenType.STRING, "aab"),
            ],
        )

    def test_comments_blanked_in_reserialization(self):
        sql = "SELECT a -- secret comment\nFROM t"
        self.assertEqual(render(sql, anonymize(sql)), "SELECT a -- ...... .......\nFROM b")

    def test_huge_numeric_literal_no_crash(self):
        tokens = anonymize("SELECT " + "1" * 4500)
        self.assertEqual(tokens[1].token_type, TokenType.NUMBER)
        self.assertEqual(len(tokens[1].text), 4500)

    def test_known_function_kept_when_passing_tokens(self):
        tokens = anonymize(
            Tokenizer(dialect="snowflake").tokenize("SELECT TO_VARIANT(x) FROM t"), "snowflake"
        )
        self.assertEqual(
            [(t.token_type, t.text) for t in tokens],
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "TO_VARIANT"),  # dialect-specific known function — kept
                (TokenType.L_PAREN, "("),
                (TokenType.VAR, "a"),  # x
                (TokenType.R_PAREN, ")"),
                (TokenType.FROM, "FROM"),
                (TokenType.VAR, "b"),  # t
            ],
        )

    def test_known_function_comment_still_blanked(self):
        self.assert_anonymized(
            "SELECT sum/* secretpassword */(x)",
            [
                (TokenType.SELECT, "SELECT"),
                (TokenType.VAR, "sum", [" .............. "]),
                (TokenType.L_PAREN, "("),
                (TokenType.VAR, "a"),  # x
                (TokenType.R_PAREN, ")"),
            ],
        )

    def test_quotes_preserved(self):
        # the alias is fitted between the quotes of the source span, so a literal stays a
        # literal of the same kind and width instead of collapsing to a bare word
        cases = [
            ("SELECT 'hello' AS x", "SELECT 'aaaaa' AS b", None),
            ("SELECT 'a''b'", "SELECT 'aaaa'", None),  # the escape is content, so it's masked
            ("SELECT ''", "SELECT ''", None),
            ("SELECT N'nat', x'4141'", "SELECT N'aaa', x'aaab'", "snowflake"),
            ('SELECT "my table"', 'SELECT "aa aaaaa"', None),
            ("SELECT `back tick`", "SELECT `aaaa aaaa`", "mysql"),
            ("SELECT [brack et]", "SELECT [aaaaa aa]", "tsql"),
            ("SELECT $$her$$, $tag$body$tag$", "SELECT $$aaa$$, $tag$aaab$tag$", "postgres"),
        ]

        for sql, expected, dialect in cases:
            rendered = render(sql, anonymize(sql, dialect), dialect)
            self.assertEqual(rendered, expected)
            self.assertEqual(len(rendered), len(sql))

    def test_source_spelling_preserved(self):
        # tokens that aren't anonymized are rendered over their own span, so the spelling
        # the tokenizer normalized away (GROUP  BY -> GROUP BY) survives
        cases = [
            ("SELECT a FROM t GROUP  BY a", "SELECT a FROM b GROUP  BY a"),
            ("SELECT a ORDER   BY a", "SELECT a ORDER   BY a"),
            ("select * from window, OUT", "select * from window, OUT"),
        ]

        for sql, expected in cases:
            rendered = render(sql, anonymize(sql))
            self.assertEqual(rendered, expected)
            self.assertEqual(len(rendered), len(sql))

    def test_numbers_keep_their_shape(self):
        sql = "SELECT 1e, 1.e, 1e10, .5, 2002, 1e-5, 1E+2"
        rendered = render(sql, anonymize(sql))
        self.assertEqual(rendered, "SELECT 1e, 2.e, 3e10, .4, 1004, 6e-1, 7E+1")
        self.assertEqual(len(rendered), len(sql))

    def test_hints_anonymized(self):
        # a hint's text is the whole /*+ ... */ comment, whether the tokenizer emits it as a
        # HINT token (only in hint position) or leaves it in the gap as a plain comment
        cases = [
            (
                "SELECT /*+ INDEX(customers ssn_idx) */ a FROM t",
                "SELECT /*+ ............... ........ */ a FROM b",
                None,
            ),
            (
                "SELECT /*+ BROADCAST(`y`) */ x FROM y",
                "SELECT /*+ .............. */ a FROM b",
                "spark",
            ),
            (
                "INSERT /*+ APPEND */ INTO t VALUES (1)",
                "INSERT /*+ ...... */ INTO a VALUES (2)",
                "oracle",
            ),
            ("SELECT a /*+ INDEX(secret) */ b", "SELECT a /*+ ............. */ b", None),
            (
                "SELECT /*+ INDEX(customers)\n           MORE(ssn) */ a FROM t",
                "SELECT /*+ ................\n           ......... */ a FROM b",
                None,
            ),
        ]

        for sql, expected, dialect in cases:
            rendered = render(sql, anonymize(sql, dialect), dialect)
            self.assertEqual(rendered, expected)
            self.assertEqual(len(rendered), len(sql))

    def test_hint_does_not_desync_following_comments(self):
        # a hint's body is recorded in Token.comments but its marker isn't left in a gap, so
        # pairing gap markers with recorded bodies drifts and copies the source through
        sql = "SELECT /*+ x */ a -- password is hunter2\nFROM t"
        self.assertEqual(
            render(sql, anonymize(sql)), "SELECT /*+ . */ a -- ........ .. .......\nFROM b"
        )

    def test_comments_without_tokens(self):
        # nothing for the tokenizer to attach the comment to
        cases = [
            ("-- top secret", "-- ... ......"),
            ("/* COMMENT */", "/* ....... */"),
            ("/*", "/*"),
        ]

        for sql, expected in cases:
            rendered = render(sql, anonymize(sql))
            self.assertEqual(rendered, expected)
            self.assertEqual(len(rendered), len(sql))

    def test_tokenize_error_keeps_delimiter(self):
        # the two leading characters survive so the delimiter the tokenizer choked on is
        # still identifiable - an unterminated string reads differently to a comment
        cases = [
            ("SELECT a, 'unterminated string", "SELECT a, 'u..................", None),
            ("SELECT a, /* unterminated comment", "SELECT a, /*.....................", None),
            ('SELECT a, "unterminated ident', 'SELECT a, "u.................', None),
            ("SELECT a, $$unterminated heredoc", "SELECT a, $$....................", "postgres"),
            ("SELECT a, `unterminated backtick", "SELECT a, `u....................", "mysql"),
            ("'unterminated secret", "'u..................", None),  # nothing tokenized at all
            ("SELECT a, '", "SELECT a, '", None),  # remainder shorter than the delimiter
        ]

        for sql, expected, dialect in cases:
            rendered = render(sql, anonymize(sql, dialect), dialect)
            self.assertEqual(rendered, expected)
            self.assertEqual(len(rendered), len(sql))

    def test_comment_bodies_fully_blanked(self):
        # markers inside a comment body are body, not structure, so they're blanked too
        cases = [
            ("SELECT a /* x /* nested */ y */ b", "SELECT a /* . .. ...... .. . */ b"),
            ("SELECT a -- /* not a real block\nFROM t", "SELECT a -- .. ... . .... .....\nFROM b"),
        ]

        for sql, expected in cases:
            rendered = render(sql, anonymize(sql))
            self.assertEqual(rendered, expected)
            self.assertEqual(len(rendered), len(sql))

    def test_render_without_dialect_still_redacts(self):
        # an unrecognized marker costs readability, never redaction
        sql = "SELECT a # secret"
        self.assertEqual(render(sql, anonymize(sql, "mysql"), "mysql"), "SELECT a # ......")
        self.assertEqual(render(sql, anonymize(sql, "mysql")), "SELECT a . ......")

    def test_render(self):
        cases = [
            ("", "", None),
            (" \t\n", " \t\n", None),
            ("-- leading secret\nSELECT foo", "-- ....... ......\nSELECT aaa", None),
            (
                "SELECT /* hidden value */ foo -- trailing secret",
                "SELECT /* ...... ..... */ aaa -- ........ ......",
                None,
            ),
            (
                "SELECT foo /* first */ /* second */ FROM bar",
                "SELECT aaa /* ..... */ /* ...... */ FROM aab",
                None,
            ),
            ("SELECT foo, 'secret tail", "SELECT aaa, 's..........", None),
            ("SELECT foo // secret", "SELECT aaa // ......", "snowflake"),
            ("SELECT foo # secret", "SELECT aaa # ......", "mysql"),
        ]

        for sql, expected, dialect in cases:
            rendered = render(sql, anonymize(sql, dialect), dialect)
            self.assertEqual(rendered, expected)
            self.assertEqual(len(rendered), len(sql))



import argparse
import inspect
import json
import os
import subprocess
import sys
import tempfile
import time


# --- Query definitions ---

large_in = (
    "SELECT * FROM t WHERE x IN (" + ", ".join(f"'s{i}'" for i in range(20000)) + ")"
    " OR y IN (" + ", ".join(str(i) for i in range(20000)) + ")"
)

values = "INSERT INTO t VALUES " + ", ".join(
    "(" + ", ".join(f"'s{i}_{j}'" if j % 2 else str(i * 20 + j) for j in range(20)) + ")"
    for i in range(2000)
)

many_joins = "SELECT * FROM t0" + "".join(
    f"\nJOIN t{i} ON t{i}.id = t{i - 1}.id" for i in range(1, 200)
)

many_unions = "\nUNION ALL\n".join(f"SELECT {i} AS a, 's{i}' AS b FROM t{i}" for i in range(500))

short = "SELECT 1 AS a, CASE WHEN 1 THEN 1 WHEN 2 THEN 2 ELSE 3 END AS b, c FROM x"

deep_arithmetic = "SELECT 1+"
deep_arithmetic += "+".join(str(i) for i in range(500))
deep_arithmetic += " AS a, 2*"
deep_arithmetic += "*".join(str(i) for i in range(500))
deep_arithmetic += " AS b FROM x"

nested_subqueries = (
    "SELECT * FROM " + "".join("(SELECT * FROM " for _ in range(20)) + "t" + ")" * 20
)

many_columns = "SELECT " + ", ".join(f"c{i}" for i in range(1000)) + " FROM t"

large_case = (
    "SELECT CASE " + " ".join(f"WHEN x = {i} THEN {i}" for i in range(1000)) + " ELSE -1 END FROM t"
)

complex_where = "SELECT * FROM t WHERE " + " AND ".join(
    f"(c{i} > {i} OR c{i} LIKE '%s{i}%' OR c{i} BETWEEN {i} AND {i + 10} OR c{i} IS NULL)"
    for i in range(200)
)

many_ctes = (
    "WITH "
    + ", ".join(f"t{i} AS (SELECT {i} AS a FROM t{i - 1 if i else 'base'})" for i in range(200))
    + " SELECT * FROM t199"
)

many_windows = (
    "SELECT "
    + ", ".join(
        f"SUM(c{i}) OVER (PARTITION BY p{i % 10} ORDER BY o{i % 5}) AS w{i}" for i in range(200)
    )
    + " FROM t"
)

nested_functions = "SELECT " + "COALESCE(" * 20 + "x" + ", NULL)" * 20 + " FROM t"

large_strings = "SELECT " + ", ".join(f"'{'x' * 100}'" for i in range(500)) + " FROM t"

many_numbers = "SELECT " + ", ".join(str(i) for i in range(10000)) + " FROM t"

tpch = """
WITH "_e_0" AS (
  SELECT
    "partsupp"."ps_partkey" AS "ps_partkey",
    "partsupp"."ps_suppkey" AS "ps_suppkey",
    "partsupp"."ps_supplycost" AS "ps_supplycost"
  FROM "partsupp" AS "partsupp"
), "_e_1" AS (
  SELECT
    "region"."r_regionkey" AS "r_regionkey",
    "region"."r_name" AS "r_name"
  FROM "region" AS "region"
  WHERE
    "region"."r_name" = 'EUROPE'
)
SELECT
  "supplier"."s_acctbal" AS "s_acctbal",
  "supplier"."s_name" AS "s_name",
  "nation"."n_name" AS "n_name",
  "part"."p_partkey" AS "p_partkey",
  "part"."p_mfgr" AS "p_mfgr",
  "supplier"."s_address" AS "s_address",
  "supplier"."s_phone" AS "s_phone",
  "supplier"."s_comment" AS "s_comment"
FROM (
  SELECT
    "part"."p_partkey" AS "p_partkey",
    "part"."p_mfgr" AS "p_mfgr",
    "part"."p_type" AS "p_type",
    "part"."p_size" AS "p_size"
  FROM "part" AS "part"
  WHERE
    "part"."p_size" = 15
    AND "part"."p_type" LIKE '%BRASS'
) AS "part"
LEFT JOIN (
  SELECT
    MIN("partsupp"."ps_supplycost") AS "_col_0",
    "partsupp"."ps_partkey" AS "_u_1"
  FROM "_e_0" AS "partsupp"
  CROSS JOIN "_e_1" AS "region"
  JOIN (
    SELECT
      "nation"."n_nationkey" AS "n_nationkey",
      "nation"."n_regionkey" AS "n_regionkey"
    FROM "nation" AS "nation"
  ) AS "nation"
    ON "nation"."n_regionkey" = "region"."r_regionkey"
  JOIN (
    SELECT
      "supplier"."s_suppkey" AS "s_suppkey",
      "supplier"."s_nationkey" AS "s_nationkey"
    FROM "supplier" AS "supplier"
  ) AS "supplier"
    ON "supplier"."s_nationkey" = "nation"."n_nationkey"
    AND "supplier"."s_suppkey" = "partsupp"."ps_suppkey"
  GROUP BY
    "partsupp"."ps_partkey"
) AS "_u_0"
  ON "part"."p_partkey" = "_u_0"."_u_1"
CROSS JOIN "_e_1" AS "region"
JOIN (
  SELECT
    "nation"."n_nationkey" AS "n_nationkey",
    "nation"."n_name" AS "n_name",
    "nation"."n_regionkey" AS "n_regionkey"
  FROM "nation" AS "nation"
) AS "nation"
  ON "nation"."n_regionkey" = "region"."r_regionkey"
JOIN "_e_0" AS "partsupp"
  ON "part"."p_partkey" = "partsupp"."ps_partkey"
JOIN (
  SELECT
    "supplier"."s_suppkey" AS "s_suppkey",
    "supplier"."s_name" AS "s_name",
    "supplier"."s_address" AS "s_address",
    "supplier"."s_nationkey" AS "s_nationkey",
    "supplier"."s_phone" AS "s_phone",
    "supplier"."s_acctbal" AS "s_acctbal",
    "supplier"."s_comment" AS "s_comment"
  FROM "supplier" AS "supplier"
) AS "supplier"
  ON "supplier"."s_nationkey" = "nation"."n_nationkey"
  AND "supplier"."s_suppkey" = "partsupp"."ps_suppkey"
WHERE
  "partsupp"."ps_supplycost" = "_u_0"."_col_0"
  AND NOT "_u_0"."_u_1" IS NULL
ORDER BY
  "supplier"."s_acctbal" DESC,
  "nation"."n_name",
  "supplier"."s_name",
  "part"."p_partkey"
LIMIT 100
"""

QUERIES = {
    "tpch": tpch,
    "short": short,
    "deep_arithmetic": deep_arithmetic,
    "large_in": large_in,
    "values": values,
    "many_joins": many_joins,
    "many_unions": many_unions,
    "nested_subqueries": nested_subqueries,
    "many_columns": many_columns,
    "large_case": large_case,
    "complex_where": complex_where,
    "many_ctes": many_ctes,
    "many_windows": many_windows,
    "nested_functions": nested_functions,
    "large_strings": large_strings,
    "many_numbers": many_numbers,
}


# --- Parser definitions ---


def sqlglot_parse(sql):
    import sqlglot

    sqlglot.parse_one(sql, error_level=sqlglot.ErrorLevel.IGNORE)


def sqlglot_transpile(sql):
    import sqlglot

    sqlglot.transpile(sql, error_level=sqlglot.ErrorLevel.IGNORE)[0]


def sqltree_parse(sql):
    import sqltree

    sqltree.api.sqltree(sql.replace('"', "`").replace("''", '"'))


def sqlparse_parse(sql):
    import sqlparse

    sqlparse.parse(sql)


def moz_sql_parser_parse(sql):
    import moz_sql_parser

    moz_sql_parser.parse(sql)


def sqloxide_parse(sql):
    import sqloxide

    sqloxide.parse_sql(sql, dialect="ansi")


def sqlfluff_parse(sql):
    import sqlfluff

    sqlfluff.parse(sql)


def polyglot_sql_parse(sql):
    import polyglot_sql

    polyglot_sql.parse_one(sql)


def polyglot_sql_transpile(sql):
    import polyglot_sql

    polyglot_sql.transpile(sql)[0]


THIRD_PARTY_PARSERS = {
    "sqltree": sqltree_parse,
    "sqlparse": sqlparse_parse,
    "sqlfluff": sqlfluff_parse,
    "moz_sql_parser": moz_sql_parser_parse,
    "sqloxide": sqloxide_parse,
    "polyglot_sql": polyglot_sql_parse,
}

THIRD_PARTY_TRANSPILERS = {
    "polyglot_sql": polyglot_sql_transpile,
}

DISPLAY_NAMES = {
    "sqlglot": "sqlglot",
    "sqlglotc": "sqlglot[c]",
    "polyglot_sql": "polyglot-sql",
    "sqltree": "sqltree",
    "sqlparse": "sqlparse",
    "moz_sql_parser": "moz_sql_parser",
    "sqlfluff": "sqlfluff",
    "sqloxide": "sqloxide",
}


# --- Third-party parser discovery ---


def _check_parser(parse_fn, queries):
    """Check which queries a parser can handle, one subprocess per query (isolates segfaults).
    Returns None if not installed, else set of query names."""
    fn_name = parse_fn.__name__
    source = inspect.getsource(parse_fn)
    supported = set()
    installed = None

    for name, sql in queries.items():
        code = f"""import signal

def _timeout(signum, frame):
    raise TimeoutError()

signal.signal(signal.SIGALRM, _timeout)
signal.alarm(5)

{source}

{fn_name}({repr(sql)})
"""
        with tempfile.NamedTemporaryFile(mode="w", encoding="utf8", suffix=".py", delete=True) as f:
            f.write(code)
            f.flush()
            try:
                result = subprocess.run([sys.executable, f.name], capture_output=True, timeout=10)
            except subprocess.TimeoutExpired:
                installed = True
                continue
            if b"ModuleNotFoundError" in result.stderr:
                return None
            installed = True
            if result.returncode == 0:
                supported.add(name)

    return supported if installed else None


def _discover_parsers(registry=None):
    """Discover available third-party parsers and which queries they support."""
    if registry is None:
        registry = THIRD_PARTY_PARSERS
    valid_pairs = set()
    available = []
    for parser_name, parse_fn in registry.items():
        supported = _check_parser(parse_fn, QUERIES)
        if supported is None:
            continue
        for query_name in supported:
            valid_pairs.add((parser_name, query_name))
        available.append(parser_name)
    return available, valid_pairs


# --- Benchmarking ---


_quiet = False


def _bench(name, fn, *args, iterations=5):
    """Benchmark fn(*args) and return the best time in seconds."""
    best = float("inf")
    for _ in range(iterations):
        t0 = time.perf_counter()
        fn(*args)
        elapsed = time.perf_counter() - t0
        if elapsed < best:
            best = elapsed
        if elapsed > 1:
            break
    if not _quiet:
        print(f"  {name}: {_fmt_time(best)}")
    return best


def _bench_sqlglot(results, mode="parse"):
    """Benchmark sqlglot (or sqlglotc if .so loaded) and add to results."""
    import sqlglot.expressions.core as _ec

    prefix = "sqlglotc" if _ec.__file__.endswith(".so") else "sqlglot"
    fn = sqlglot_transpile if mode == "transpile" else sqlglot_parse
    for query_name, sql in QUERIES.items():
        key = f"{prefix}:{query_name}"
        results[key] = _bench(key, fn, sql)
    return prefix


def _bench_third_party(results, mode="parse"):
    """Benchmark third-party parsers/transpilers and add to results. Returns list of available names."""
    registry = THIRD_PARTY_TRANSPILERS if mode == "transpile" else THIRD_PARTY_PARSERS
    available, valid_pairs = _discover_parsers(registry)
    for query_name, sql in QUERIES.items():
        for name, fn in registry.items():
            if (name, query_name) in valid_pairs:
                key = f"{name}:{query_name}"
                results[key] = _bench(key, fn, sql)
    return available


# --- Table printing ---


def _fmt_ratio(ratio):
    return f"{ratio:.2f}"


def _fmt_time(seconds):
    if seconds >= 1:
        return f"{seconds:.2f} sec"
    if seconds >= 1e-3:
        return f"{seconds * 1e3:.2f} ms"
    return f"{seconds * 1e6:.1f} us"


def _print_table(base_parser, all_parsers, results):
    query_width = max(len(q) for q in QUERIES)
    query_width = max(query_width, len("Query"))

    # Pre-compute all cells to determine column widths
    cells = {}
    for query_name in QUERIES:
        base_time = results.get(f"{base_parser}:{query_name}")
        for p in all_parsers:
            t = results.get(f"{p}:{query_name}")
            if t is not None and base_time:
                ratio = t / base_time
                cells[(p, query_name)] = f"{t:.6f} ({_fmt_ratio(ratio)})"
            else:
                cells[(p, query_name)] = "N/A"

    col_widths = {}
    for p in all_parsers:
        name = DISPLAY_NAMES.get(p, p)
        w = len(name)
        for query_name in QUERIES:
            w = max(w, len(cells[(p, query_name)]))
        col_widths[p] = w

    header = f"| {'Query':>{query_width}} |"
    sep = f"| {'-' * query_width} |"
    for p in all_parsers:
        name = DISPLAY_NAMES.get(p, p)
        header += f" {name:>{col_widths[p]}} |"
        sep += f" {'-' * col_widths[p]} |"

    print()
    print(header)
    print(sep)

    for query_name in QUERIES:
        row = f"| {query_name:>{query_width}} |"
        for p in all_parsers:
            row += f" {cells[(p, query_name)]:>{col_widths[p]}} |"
        print(row)


# --- Subprocess entry point for .so mode ---


def _has_so_files():
    import glob

    return bool(glob.glob("sqlglot/**/*.so", recursive=True))


def _run_subprocess():
    """Run sqlglot benchmarks and print results to stdout as key=value lines."""
    global _quiet
    _quiet = bool(os.environ.get("_BENCH_QUIET"))
    mode = os.environ.get("_BENCH_MODE", "parse")
    results = {}
    _bench_sqlglot(results, mode=mode)
    for key, value in results.items():
        print(f"{key}={value}")


# --- Main ---


def _parse_args():
    parser = argparse.ArgumentParser(description="SQLGlot parser benchmarks")
    parser.add_argument("--json", metavar="FILE", help="Write results as JSON to FILE")
    parser.add_argument("--quiet", action="store_true", help="Suppress progress output")
    parser.add_argument(
        "--sqlglot-only",
        action="store_true",
        help="Only benchmark sqlglot/sqlglotc (skip third-party parsers)",
    )
    parser.add_argument(
        "--mode",
        choices=["parse", "transpile"],
        default="parse",
        help="Benchmark mode: parse or transpile (default: parse)",
    )
    return parser.parse_args()


if __name__ == "__main__":
    if os.environ.get("_BENCH_SUBPROCESS"):
        _run_subprocess()
    else:
        args = _parse_args()
        _quiet = args.quiet

        mode = args.mode

        if _has_so_files():
            if not _quiet:
                print("=== Running sqlglot[c] ===", flush=True)
            env = {**os.environ, "_BENCH_SUBPROCESS": "1", "_BENCH_MODE": mode}
            if _quiet:
                env["_BENCH_QUIET"] = "1"
            proc = subprocess.run(
                [sys.executable, __file__], env=env, capture_output=True, text=True, check=True
            )
            results = {}
            for line in proc.stdout.splitlines():
                if "=" in line:
                    key, value = line.split("=", 1)
                    results[key] = float(value)
                elif not _quiet:
                    print(line)

            if not _quiet:
                print("\n=== Hiding .so files ===", flush=True)
            subprocess.run(["make", "hidec"], check=True, capture_output=True)

            try:
                if not _quiet:
                    print("\n=== Running pure Python ===", flush=True)
                _bench_sqlglot(results, mode=mode)
                if not args.sqlglot_only:
                    if not _quiet:
                        print(f"\n=== Running third-party ({mode}) ===", flush=True)
                    available = _bench_third_party(results, mode=mode)
                else:
                    available = []
            finally:
                subprocess.run(["make", "showc"], capture_output=True)

            if args.json:
                with open(args.json, "w") as f:
                    json.dump(results, f, indent=2)
            else:
                _print_table("sqlglot", ["sqlglot", "sqlglotc"] + available, results)
        else:
            results = {}
            prefix = _bench_sqlglot(results, mode=mode)
            if not args.sqlglot_only:
                available = _bench_third_party(results, mode=mode)
            else:
                available = []

            if args.json:
                with open(args.json, "w") as f:
                    json.dump(results, f, indent=2)
            else:
                _print_table(prefix, [prefix] + available, results)



from __future__ import annotations

from sqlglot import exp
from sqlglot.generators.trino import TrinoGenerator


class DuneGenerator(TrinoGenerator):
    TRY_SUPPORTED = False
    SUPPORTS_UESCAPE = False

    TRANSFORMS = {
        **TrinoGenerator.TRANSFORMS,
        exp.HexString: lambda self, e: f"0x{e.this}",
    }



from __future__ import annotations

import typing as t

from sqlglot import expressions as exp
from sqlglot import generator
from sqlglot.dialects.dialect import (
    rename_func,
    no_trycast_sql,
)

DATE_DELTA = t.Union[exp.DateAdd, exp.DateSub]


def _date_delta_sql(name: str) -> t.Callable[[DremioGenerator, DATE_DELTA], str]:
    def _delta_sql(self: DremioGenerator, expression: DATE_DELTA) -> str:
        unit = expression.text("unit").upper()

        # Fallback to default behavior if unit is missing or 'DAY'
        if not unit or unit == "DAY":
            return self.func(name, expression.this, expression.expression)

        this_sql = self.sql(expression, "this")
        expr_sql = self.sql(expression, "expression")

        interval_sql = f"CAST({expr_sql} AS INTERVAL {unit})"
        return f"{name}({this_sql}, {interval_sql})"

    return _delta_sql


class DremioGenerator(generator.Generator):
    SELECT_KINDS: tuple[str, ...] = ()
    TRY_SUPPORTED = False
    SUPPORTS_UESCAPE = False
    SUPPORTS_DECODE_CASE = False

    AFTER_HAVING_MODIFIER_TRANSFORMS = generator.AFTER_HAVING_MODIFIER_TRANSFORMS

    NVL2_SUPPORTED = False
    SUPPORTS_CONVERT_TIMEZONE = True
    INTERVAL_ALLOWS_PLURAL_FORM = False
    JOIN_HINTS = False
    LIMIT_ONLY_LITERALS = True
    MULTI_ARG_DISTINCT = False
    SUPPORTS_BETWEEN_FLAGS = True

    # https://docs.dremio.com/current/reference/sql/data-types/
    TYPE_MAPPING = {
        **generator.Generator.TYPE_MAPPING,
        exp.DType.SMALLINT: "INT",
        exp.DType.TINYINT: "INT",
        exp.DType.BINARY: "VARBINARY",
        exp.DType.TEXT: "VARCHAR",
        exp.DType.NCHAR: "VARCHAR",
        exp.DType.CHAR: "VARCHAR",
        exp.DType.TIMESTAMPNTZ: "TIMESTAMP",
        exp.DType.DATETIME: "TIMESTAMP",
        exp.DType.ARRAY: "LIST",
        exp.DType.BIT: "BOOLEAN",
    }

    TRANSFORMS = {
        **generator.Generator.TRANSFORMS,
        exp.BitwiseAndAgg: rename_func("BIT_AND"),
        exp.BitwiseOrAgg: rename_func("BIT_OR"),
        exp.ToChar: rename_func("TO_CHAR"),
        exp.TimeToStr: lambda self, e: self.func("TO_CHAR", e.this, self.format_time(e)),
        exp.TryCast: no_trycast_sql,
        exp.DateAdd: _date_delta_sql("DATE_ADD"),
        exp.DateSub: _date_delta_sql("DATE_SUB"),
        exp.GenerateSeries: rename_func("ARRAY_GENERATE_RANGE"),
    }

    def version_sql(self, expression: exp.Version) -> str:
        if expression.text("kind") == "AS OF":
            this = "SNAPSHOT" if expression.name == "VERSION" else expression.name
            return f"AT {this} {self.sql(expression, 'expression')}"

        self.unsupported("Range time travel is not supported in Dremio")
        return super().version_sql(expression)

    def datatype_sql(self, expression: exp.DataType) -> str:
        """
        Reject time-zone-aware TIMESTAMPs, which Dremio does not accept
        """
        if expression.is_type(
            exp.DType.TIMESTAMPTZ,
            exp.DType.TIMESTAMPLTZ,
        ):
            self.unsupported("Dremio does not support time-zone-aware TIMESTAMP")

        return super().datatype_sql(expression)

    def cast_sql(self, expression: exp.Cast, safe_prefix: str | None = None) -> str:
        # Match: CAST(CURRENT_TIMESTAMP AT TIME ZONE 'UTC' AS DATE)
        if expression.is_type(exp.DType.DATE):
            at_time_zone = expression.this

            if (
                isinstance(at_time_zone, exp.AtTimeZone)
                and isinstance(at_time_zone.this, exp.CurrentTimestamp)
                and isinstance(at_time_zone.args["zone"], exp.Literal)
                and at_time_zone.text("zone").upper() == "UTC"
            ):
                return "CURRENT_DATE_UTC"

        return super().cast_sql(expression, safe_prefix)



from sqlglot import tokens
from sqlglot.dialects.dialect import Dialect, NormalizationStrategy
from sqlglot.generators.solr import SolrGenerator
from sqlglot.parsers.solr import SolrParser


# https://solr.apache.org/guide/solr/latest/query-guide/sql-query.html


class Solr(Dialect):
    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
    DPIPE_IS_STRING_CONCAT = False

    Generator = SolrGenerator

    Parser = SolrParser

    class Tokenizer(tokens.Tokenizer):
        QUOTES = ["'"]
        IDENTIFIERS = ["`"]



from __future__ import annotations

import json
import logging
import typing as t
from dataclasses import dataclass, field

from sqlglot import Schema, exp, maybe_parse
from sqlglot.errors import SqlglotError
from sqlglot.optimizer import Scope, build_scope, find_all_in_scope, normalize_identifiers, qualify
from sqlglot.optimizer.scope import ScopeType
from sqlglot.schema import ensure_schema

if t.TYPE_CHECKING:
    from sqlglot.dialects.dialect import DialectType
    from collections.abc import Iterator, Mapping, Sequence
    from sqlglot._typing import GraphHTMLArgs
    from typing_extensions import Unpack

logger = logging.getLogger("sqlglot")


@dataclass(frozen=True)
class Node:
    name: str
    expression: exp.Expr
    source: exp.Expr
    downstream: list[Node] = field(default_factory=list)
    source_name: str = ""
    reference_node_name: str = ""

    # Caller-injected per-node data, populated via the `on_node` hook on lineage()
    payload: dict[str, t.Any] = field(default_factory=dict)

    def walk(self) -> Iterator[Node]:
        visited: set[int] = set()
        queue = [self]
        while queue:
            node = queue.pop()
            node_id = id(node)
            if node_id in visited:
                continue
            visited.add(node_id)
            yield node
            queue.extend(reversed(node.downstream))

    def to_html(self, dialect: DialectType = None, **opts: Unpack[GraphHTMLArgs]) -> GraphHTML:
        nodes = {}
        edges = []

        for node in self.walk():
            if isinstance(node.expression, exp.Table):
                label = f"FROM {node.expression.this}"
                title = f"<pre>SELECT {node.name} FROM {node.expression.this}</pre>"
                group = 1
            else:
                label = node.expression.sql(pretty=True, dialect=dialect)
                source = node.source.transform(
                    lambda n: (
                        exp.Tag(this=n, prefix="<b>", postfix="</b>") if n is node.expression else n
                    ),
                    copy=False,
                ).sql(pretty=True, dialect=dialect)
                title = f"<pre>{source}</pre>"
                group = 0

            node_id = id(node)

            nodes[node_id] = {
                "id": node_id,
                "label": label,
                "title": title,
                "group": group,
            }

            for d in node.downstream:
                edges.append({"from": node_id, "to": id(d)})
        return GraphHTML(nodes, edges, **opts)


@t.overload
def lineage(column: str | exp.Column, sql: str | exp.Expr, **kwargs: t.Any) -> Node: ...


@t.overload
def lineage(column: None, sql: str | exp.Expr, **kwargs: t.Any) -> dict[str, Node]: ...


def lineage(
    column: str | exp.Column | None,
    sql: str | exp.Expr,
    schema: dict | Schema | None = None,
    sources: Mapping[str, str | exp.Query] | None = None,
    dialect: DialectType = None,
    scope: Scope | None = None,
    trim_selects: bool = True,
    copy: bool = True,
    on_node: t.Callable[[Node], None] | None = None,
    **kwargs,
) -> Node | dict[str, Node]:
    """Build the lineage graph for a SQL query.

    If `column` is given, returns the lineage Node for that single output column.
    If `column` is None, returns a dict mapping every top-level output column name
    to its lineage Node (with a shared cache so cross-column work is deduplicated).

    Args:
        column: The column to build the lineage for. Pass None to get all output columns.
        sql: The SQL string or expression.
        schema: The schema of tables.
        sources: A mapping of queries which will be used to continue building lineage.
        dialect: The dialect of input SQL.
        scope: A pre-created scope to use instead.
        trim_selects: Whether to clean up selects by trimming to only relevant columns.
        copy: Whether to copy the Expr arguments.
        on_node: Optional callback invoked for every Node created during the walk,
            after the Node's downstream is populated. Useful for injecting
            caller-managed data into Node.payload during the walk.
        **kwargs: Qualification optimizer kwargs.

    Returns:
        A Node when `column` is provided, or a dict[str, Node] when `column` is None.
    """
    expression = maybe_parse(sql, copy=copy, dialect=dialect)

    if sources:
        expression = exp.expand(
            expression,
            {
                k: t.cast(exp.Query, maybe_parse(v, copy=copy, dialect=dialect))
                for k, v in sources.items()
            },
            dialect=dialect,
            copy=copy,
        )

    schema = ensure_schema(schema, dialect=dialect)

    if not scope:
        expression = qualify.qualify(
            expression,
            dialect=dialect,
            schema=schema,
            **{"validate_qualify_columns": False, "identify": False, **kwargs},  # type: ignore
        )
        scope = build_scope(expression)

    if not scope:
        raise SqlglotError("Cannot build lineage, sql must be SELECT")

    selectable = scope.expression
    if not isinstance(selectable, exp.Selectable):
        raise SqlglotError("Cannot build lineage, sql must be a query")

    cache: dict[tuple, Node] = {}
    scope_meta: dict[int, tuple[bool, dict[str, exp.Expr]]] = {}

    if column is not None:
        column_name = normalize_identifiers.normalize_identifiers(column, dialect=dialect).name
        if not any(select.alias_or_name == column_name for select in selectable.selects):
            raise SqlglotError(f"Cannot find column '{column_name}' in query.")

        return to_node(
            column_name,
            scope,
            dialect,
            trim_selects=trim_selects,
            schema=schema,
            _cache=cache,
            _scope_meta=scope_meta,
            on_node=on_node,
        )

    result: dict[str, Node] = {}
    for sel in selectable.selects:
        name = sel.alias_or_name
        if not name:
            raise SqlglotError(
                f"Cannot fetch lineage for unnamed projection: {sel.sql(dialect=dialect)}."
            )

        result[name] = to_node(
            name,
            scope,
            dialect,
            trim_selects=trim_selects,
            schema=schema,
            _cache=cache,
            _scope_meta=scope_meta,
            on_node=on_node,
        )

    return result


def to_node(
    column: str | int,
    scope: Scope,
    dialect: DialectType,
    scope_name: str | None = None,
    upstream: Node | None = None,
    source_name: str | None = None,
    reference_node_name: str | None = None,
    trim_selects: bool = True,
    schema: Schema | None = None,
    _cache: dict[tuple, Node] | None = None,
    _scope_meta: dict[int, tuple[bool, dict[str, exp.Expr]]] | None = None,
    on_node: t.Callable[[Node], None] | None = None,
) -> Node:
    cache_key = (column, id(scope), scope_name, source_name, reference_node_name)

    if _cache is not None and cache_key in _cache:
        cached_node = _cache[cache_key]
        if upstream:
            upstream.downstream.append(cached_node)
        return cached_node

    # Find the specific select clause that is the source of the column we want.
    # This can either be a specific, named select or a generic `*` clause.
    selectable = t.cast(exp.Selectable, scope.expression)
    if isinstance(column, int):
        if column >= len(selectable.selects):
            raise SqlglotError(
                f"Cannot find column's source with index {column} in query: {selectable.sql(dialect=dialect)}"
            )
        select = selectable.selects[column]
    else:
        # Resolving a column to its select scans selectable.selects on every call;
        # memoize a per-scope {name: select} map and is_star bit instead.
        if _scope_meta is None:
            select = next(
                (s for s in selectable.selects if s.alias_or_name == column),
                exp.Star() if selectable.is_star else scope.expression,
            )
        else:
            scope_id = id(scope)
            meta = _scope_meta.get(scope_id)
            if meta is None:
                select_by_name: dict[str, exp.Expr] = {}
                for sel in selectable.selects:
                    select_by_name.setdefault(sel.alias_or_name, sel)
                meta = (selectable.is_star, select_by_name)
                _scope_meta[scope_id] = meta
            is_star, select_by_name = meta
            select = select_by_name.get(column, exp.Star() if is_star else scope.expression)

    if isinstance(scope.expression, exp.Subquery):
        for inner_scope in scope.subquery_scopes:
            result = to_node(
                column,
                scope=inner_scope,
                dialect=dialect,
                upstream=upstream,
                source_name=source_name,
                reference_node_name=reference_node_name,
                trim_selects=trim_selects,
                schema=schema,
                _cache=_cache,
                _scope_meta=_scope_meta,
                on_node=on_node,
            )
            # Skip caching a passed-in upstream returned by an inner SetOp:
            # a sibling call at the same key with that node as its upstream
            # would otherwise self-loop on the cache hit.
            if _cache is not None and result is not upstream:
                _cache[cache_key] = result
            return result
    if isinstance(scope.expression, exp.SetOperation):
        name = type(scope.expression).__name__.upper()
        created_setop = upstream is None
        upstream = upstream or Node(name=name, source=scope.expression, expression=select)

        index = (
            column
            if isinstance(column, int)
            else next(
                (
                    i
                    for i, select in enumerate(selectable.selects)
                    if select.alias_or_name == column or select.is_star
                ),
                -1,  # mypy will not allow a None here, but a negative index should never be returned
            )
        )

        if index == -1:
            raise ValueError(f"Could not find {column} in {scope.expression}")

        for s in scope.union_scopes:
            to_node(
                index,
                scope=s,
                dialect=dialect,
                upstream=upstream,
                source_name=source_name,
                reference_node_name=reference_node_name,
                trim_selects=trim_selects,
                schema=schema,
                _cache=_cache,
                _scope_meta=_scope_meta,
                on_node=on_node,
            )

        if _cache is not None and created_setop:
            _cache[cache_key] = upstream
        if created_setop and on_node:
            on_node(upstream)
        return upstream

    if trim_selects and isinstance(scope.expression, exp.Select):
        # For better ergonomics in our node labels, replace the full select with
        # a version that has only the column we care about.
        #   "x", SELECT x, y FROM foo
        #     => "x", SELECT x FROM foo
        source: exp.Expr = scope.expression.select(select, append=False)
    else:
        source = scope.expression

    # Create the node for this step in the lineage chain, and attach it to the previous one.
    node = Node(
        name=f"{scope_name}.{column}" if scope_name else str(column),
        source=source,
        expression=select,
        source_name=source_name or "",
        reference_node_name=reference_node_name or "",
    )

    if upstream:
        upstream.downstream.append(node)

    subquery_scopes = {
        id(subquery_scope.expression): subquery_scope for subquery_scope in scope.subquery_scopes
    }

    for subquery in find_all_in_scope(select, *exp.UNWRAPPED_QUERIES):
        subquery_scope: Scope | None = subquery_scopes.get(id(subquery))
        if not subquery_scope:
            logger.warning(f"Unknown subquery scope: {subquery.sql(dialect=dialect)}")
            continue

        for name in subquery.named_selects:
            to_node(
                name,
                scope=subquery_scope,
                dialect=dialect,
                upstream=node,
                trim_selects=trim_selects,
                schema=schema,
                _cache=_cache,
                _scope_meta=_scope_meta,
                on_node=on_node,
            )

    # if the select is a star add all scope sources as downstreams
    if isinstance(select, exp.Star):
        for src in scope.sources.values():
            src_expr = src.expression if isinstance(src, Scope) else src
            star_node = Node(name=select.sql(comments=False), source=src_expr, expression=src_expr)
            node.downstream.append(star_node)
            if on_node:
                on_node(star_node)

    # Find all columns that went into creating this one to list their lineage nodes.
    source_columns = set(find_all_in_scope(select, exp.Column))

    # If the source is a UDTF find columns used in the UDTF to generate the table
    if isinstance(source, exp.UDTF):
        source_columns |= set(source.find_all(exp.Column))
        derived_tables: Sequence[exp.Expr] = [
            src.expression.parent
            for src in scope.sources.values()
            if isinstance(src, Scope) and src.is_derived_table and src.expression.parent
        ]
    else:
        derived_tables = scope.derived_tables

    source_names = {
        dt.alias: dt.comments[0].split()[1]
        for dt in derived_tables
        if dt.comments and dt.comments[0].startswith("source: ")
    }

    pivots = scope.pivots
    if pivots and pivots[0].parent is not pivots[-1].parent:
        # The scope's pivots only form a chain when they all hang off of the same source;
        # otherwise they can't be folded, so their columns degrade to unresolved leaves
        pivots = []

    pivot_renames: dict[str, str] = {}
    pivot_column_mapping: dict[str, list[exp.Column]] = {}

    if pivots:
        pivot_renames, pivot_column_mapping = _pivot_chain_mapping(pivots, scope, schema)

    for c in source_columns:
        table = c.table
        col_source: exp.Table | Scope | None = scope.sources.get(table)

        if isinstance(col_source, Scope):
            reference_node_name = None
            if col_source.scope_type == ScopeType.DERIVED_TABLE and table not in source_names:
                reference_node_name = table
            elif col_source.scope_type == ScopeType.CTE:
                selected_node, _ = scope.selected_sources.get(table, (None, None))
                reference_node_name = selected_node.name if selected_node else None

            # The table itself came from a more specific scope. Recurse into that one using the unaliased column name.
            to_node(
                c.name,
                scope=col_source,
                dialect=dialect,
                scope_name=table,
                upstream=node,
                source_name=source_names.get(table) or source_name,
                reference_node_name=reference_node_name,
                trim_selects=trim_selects,
                schema=schema,
                _cache=_cache,
                _scope_meta=_scope_meta,
                on_node=on_node,
            )
        elif pivots and pivots[-1].alias_or_name == c.table:
            # Only the last operator in a chain names the resulting source
            pivot_parent = pivots[-1].parent
            downstream_columns = []

            column_name = c.name
            if column_name in pivot_column_mapping:
                downstream_columns.extend(pivot_column_mapping[column_name])
            else:
                # The column is not in the pivot, so it must be an implicit column of the
                # pivoted source -- adapt column to be from the implicit pivoted source.
                downstream_columns.append(
                    exp.column(
                        pivot_renames.get(c.name, c.this),
                        table=pivot_parent.alias_or_name if pivot_parent else None,
                    )
                )

            for downstream_column in downstream_columns:
                if not downstream_column.table:
                    # Some dialects (e.g. bigquery) don't qualify the IN-list columns,
                    # but they can only come from the pivoted source
                    downstream_column = exp.column(
                        downstream_column.this,
                        table=pivot_parent.alias_or_name if pivot_parent else None,
                    )

                table = downstream_column.table
                col_source = scope.sources.get(table)
                if isinstance(col_source, exp.Table) and not col_source.db:
                    # A pivoted CTE reference maps to the raw table in `scope.sources`,
                    # so recover the CTE's scope to keep tracing through it
                    col_source = scope.cte_sources.get(col_source.name, col_source)
                if isinstance(col_source, Scope):
                    to_node(
                        downstream_column.name,
                        scope=col_source,
                        scope_name=table,
                        dialect=dialect,
                        upstream=node,
                        source_name=source_names.get(table) or source_name,
                        reference_node_name=reference_node_name,
                        trim_selects=trim_selects,
                        schema=schema,
                        _cache=_cache,
                        _scope_meta=_scope_meta,
                        on_node=on_node,
                    )
                else:
                    col_expr = col_source or exp.Placeholder()
                    pivot_leaf = Node(
                        name=downstream_column.sql(comments=False),
                        source=col_expr,
                        expression=col_expr,
                    )
                    node.downstream.append(pivot_leaf)
                    if on_node:
                        on_node(pivot_leaf)
        else:
            # The source is not a scope and the column is not in any pivot - we've reached the end
            # of the line. At this point, if a source is not found it means this column's lineage
            # is unknown. This can happen if the definition of a source used in a query is not
            # passed into the `sources` map.
            col_expr = col_source or exp.Placeholder()
            leaf = Node(name=c.sql(comments=False), source=col_expr, expression=col_expr)
            node.downstream.append(leaf)
            if on_node:
                on_node(leaf)

    if _cache is not None:
        _cache[cache_key] = node

    if on_node:
        on_node(node)

    return node


def _pre_pivot_columns(pivot: exp.Pivot, scope: Scope, schema: Schema | None = None) -> list[str]:
    """
    The columns the first operator of a chain sees, taken from the projections of a
    derived table or CTE source, or from the schema for a physical table. Returns an
    empty list when they can't be determined (e.g. an unexpanded star), since anything
    positional over them would silently shift.
    """
    parent = pivot.parent
    columns: list[str] = []
    if isinstance(parent, exp.DerivedTable) and isinstance(parent.this, exp.Query):
        columns = parent.this.named_selects
    elif isinstance(parent, exp.Table):
        cte_source = scope.cte_sources.get(parent.name) if not parent.db else None
        if isinstance(cte_source, Scope) and isinstance(cte_source.expression, exp.Query):
            columns = cte_source.expression.named_selects
        elif schema is not None:
            columns = list(schema.column_names(parent, only_visible=True))

    return [] if "*" in columns else columns


def _pivot_chain_mapping(
    pivots: list[exp.Pivot], scope: Scope, schema: Schema | None = None
) -> tuple[dict[str, str], dict[str, list[exp.Column]]]:
    """
    Fold a chain of (UN)PIVOT operators into a single view of its output, since each one
    consumes the previous one's columns rather than the pivoted source's.

    Returns the composed output-name -> pre-chain-name renames (from alias column lists),
    and the composed output-name -> source columns it derives from.
    """
    available = _pre_pivot_columns(pivots[0], scope, schema)
    renames: dict[str, str] = {}
    mapping: dict[str, list[exp.Column]] = {}

    for pivot in pivots:
        # Renames are positional over the operator's full output, so they can only be
        # applied when the columns going into it are known
        step_renames = (
            pivot.output_columns(available) if pivot.alias_column_names and available else {}
        )
        step_mapping = _pivot_column_mapping(pivot)
        if step_renames:
            step_mapping = {
                post: step_mapping[pre] for post, pre in step_renames.items() if pre in step_mapping
            }

        # Columns this operator consumed may have been produced by an earlier one, or be
        # alias-list renames of passthroughs; resolve through what we've folded so far
        def resolve(col: exp.Column) -> list[exp.Column]:
            if col.name in mapping:
                return mapping[col.name]
            if col.name in renames:
                return [exp.column(renames[col.name])]
            return [col]

        composed = {
            out: [resolved for col in cols for resolved in resolve(col)]
            for out, cols in step_mapping.items()
        }

        # Whatever an earlier operator produced and this one didn't consume passes through,
        # under whatever name this operator's alias column list gives it
        consumed = {col.name for cols in step_mapping.values() for col in cols}
        pre_to_post = {pre: post for post, pre in step_renames.items()}
        for out, cols in mapping.items():
            if out not in consumed:
                composed.setdefault(pre_to_post.get(out, out), cols)

        renames = (
            {post: renames.get(pre, pre) for post, pre in step_renames.items()}
            if step_renames
            else renames
        )
        mapping = composed
        available = list(pivot.output_columns(available)) if available else []

    return renames, mapping


def _pivot_column_mapping(pivot: exp.Pivot) -> dict[str, list[exp.Column]]:
    """Map each (UN)PIVOT output column name to the source columns it's derived from."""
    mapping: dict[str, list[exp.Column]] = {}

    if pivot.unpivot:
        # UNPIVOT((v1, v2) FOR name IN ((a1, a2), (b1, b2))): each value column is derived
        # positionally from the IN-list entries, and the name column from all of them
        value_columns = [
            identifier for e in pivot.expressions for identifier in e.find_all(exp.Identifier)
        ]
        for value_column in value_columns:
            mapping[value_column.name] = []

        for field in pivot.fields:
            if not isinstance(field, exp.In):
                continue

            name_columns = mapping.setdefault(field.this.name, [])
            for entry in field.expressions:
                entry_columns = list(entry.find_all(exp.Column))
                name_columns.extend(entry_columns)

                if len(entry_columns) == len(value_columns):
                    for value_column, column in zip(value_columns, entry_columns):
                        mapping[value_column.name].append(column)
                else:
                    for value_column in value_columns:
                        mapping[value_column.name].extend(entry_columns)

        return mapping

    # For each aggregation function, the pivot creates a new column for each field in category
    # combined with the aggfunc. So the columns parsed have this order: cat_a_value_sum, cat_a,
    # b_value_sum, b. Because of this step wise manner the aggfunc 'sum(value) as value_sum'
    # belongs to the column indices 0, 2, and the aggfunc 'max(price)' without an alias belongs
    # to the column indices 1, 3. Here, only the columns used in the aggregations are of interest
    # in the lineage, so lookup the pivot column name by index and map that with the columns used
    # in the aggregation.
    #
    # Example: PIVOT (SUM(value) AS value_sum, MAX(price)) FOR category IN ('a' AS cat_a, 'b')
    pivot_columns = pivot.args["columns"]
    pivot_aggs_count = len(pivot.expressions)

    mapping = {}
    for i, agg in enumerate(pivot.expressions):
        agg_cols = list(agg.find_all(exp.Column))
        for col_index in range(i, len(pivot_columns), pivot_aggs_count):
            mapping[pivot_columns[col_index].name] = agg_cols
    return mapping


class GraphHTML:
    """Node to HTML generator using vis.js.

    https://visjs.github.io/vis-network/docs/network/
    """

    def __init__(
        self,
        nodes: dict,
        edges: list,
        imports: bool = True,
        options: Mapping[str, object] | None = None,
    ):
        self.imports = imports

        self.options = {
            "height": "500px",
            "width": "100%",
            "layout": {
                "hierarchical": {
                    "enabled": True,
                    "nodeSpacing": 200,
                    "sortMethod": "directed",
                },
            },
            "interaction": {
                "dragNodes": False,
                "selectable": False,
            },
            "physics": {
                "enabled": False,
            },
            "edges": {
                "arrows": "to",
            },
            "nodes": {
                "font": "20px monaco",
                "shape": "box",
                "widthConstraint": {
                    "maximum": 300,
                },
            },
            **(options or {}),
        }

        self.nodes = nodes
        self.edges = edges

    def __str__(self):
        nodes = json.dumps(list(self.nodes.values()))
        edges = json.dumps(self.edges)
        options = json.dumps(self.options)
        imports = (
            """<script type="text/javascript" src="https://unpkg.com/vis-data@latest/peer/umd/vis-data.min.js"></script>
  <script type="text/javascript" src="https://unpkg.com/vis-network@latest/peer/umd/vis-network.min.js"></script>
  <link rel="stylesheet" type="text/css" href="https://unpkg.com/vis-network/styles/vis-network.min.css" />"""
            if self.imports
            else ""
        )

        return f"""<div>
  <div id="sqlglot-lineage"></div>
  {imports}
  <script type="text/javascript">
    var nodes = new vis.DataSet({nodes})
    nodes.forEach(row => row["title"] = new DOMParser().parseFromString(row["title"], "text/html").body.childNodes[0])

    new vis.Network(
        document.getElementById("sqlglot-lineage"),
        {{
            nodes: nodes,
            edges: new vis.DataSet({edges})
        }},
        {options},
    )
  </script>
</div>"""

    def _repr_html_(self) -> str:
        return self.__str__()



from __future__ import annotations

from sqlglot import tokens
from sqlglot.dialects.dialect import Dialect
from sqlglot.generators.prql import PRQLGenerator
from sqlglot.parsers.prql import PRQLParser
from sqlglot.tokens import TokenType


class PRQL(Dialect):
    DPIPE_IS_STRING_CONCAT = False

    Generator = PRQLGenerator

    class Tokenizer(tokens.Tokenizer):
        IDENTIFIERS = ["`"]
        QUOTES = ["'", '"']

        SINGLE_TOKENS = {
            **tokens.Tokenizer.SINGLE_TOKENS,
            "=": TokenType.ALIAS,
            "'": TokenType.QUOTE,
            '"': TokenType.QUOTE,
            "`": TokenType.IDENTIFIER,
            "#": TokenType.COMMENT,
        }

        KEYWORDS = {
            **tokens.Tokenizer.KEYWORDS,
        }

    Parser = PRQLParser



from __future__ import annotations

import typing as t

from sqlglot import exp
from sqlglot.trie import new_trie
from sqlglot.dialects.dialect import build_date_delta, build_like
from sqlglot.helper import ensure_list, seq_get
from sqlglot.parsers.hive import build_with_ignore_nulls
from sqlglot.parsers.spark2 import Spark2Parser, build_as_cast
from sqlglot.tokens import TokenType


def _build_datediff(args: list) -> exp.Expr:
    """
    Although Spark docs don't mention the "unit" argument, Spark3 added support for
    it at some point. Databricks also supports this variant (see below).

    For example, in spark-sql (v3.3.1):
    - SELECT DATEDIFF('2020-01-01', '2020-01-05') results in -4
    - SELECT DATEDIFF(day, '2020-01-01', '2020-01-05') results in 4

    See also:
    - https://docs.databricks.com/sql/language-manual/functions/datediff3.html
    - https://docs.databricks.com/sql/language-manual/functions/datediff.html
    """
    unit = None
    this = seq_get(args, 0)
    expression = seq_get(args, 1)

    if len(args) == 3:
        unit = exp.var(t.cast(exp.Expr, this).name)
        this = args[2]

    return exp.DateDiff(
        this=exp.TsOrDsToDate(this=this), expression=exp.TsOrDsToDate(this=expression), unit=unit
    )


def _build_dateadd(args: list) -> exp.Expr:
    expression = seq_get(args, 1)

    if len(args) == 2:
        # DATE_ADD(startDate, numDays INTEGER)
        # https://docs.databricks.com/en/sql/language-manual/functions/date_add.html
        return exp.TsOrDsAdd(
            this=seq_get(args, 0), expression=expression, unit=exp.Literal.string("DAY")
        )

    # DATE_ADD / DATEADD / TIMESTAMPADD(unit, value integer, expr)
    # https://docs.databricks.com/en/sql/language-manual/functions/date_add3.html
    return exp.TimestampAdd(this=seq_get(args, 2), expression=expression, unit=seq_get(args, 0))


class SparkParser(Spark2Parser):
    NO_PAREN_FUNCTIONS = {
        **Spark2Parser.NO_PAREN_FUNCTIONS,
        TokenType.SESSION_USER: exp.SessionUser,
    }

    SET_PARSERS = {
        **Spark2Parser.SET_PARSERS,
        "VAR": lambda self: self._parse_set_item_assignment("VARIABLE"),
        "VARIABLE": lambda self: self._parse_set_item_assignment("VARIABLE"),
    }

    SET_TRIE = new_trie(key.split(" ") for key in SET_PARSERS)

    FUNCTIONS = {
        **Spark2Parser.FUNCTIONS,
        "ANY_VALUE": build_with_ignore_nulls(exp.AnyValue),
        "ARRAY_INSERT": lambda args: exp.ArrayInsert(
            this=seq_get(args, 0),
            position=seq_get(args, 1),
            expression=seq_get(args, 2),
            offset=1,
        ),
        "BIT_AND": exp.BitwiseAndAgg.from_arg_list,
        "BIT_GET": exp.Getbit.from_arg_list,
        "BIT_OR": exp.BitwiseOrAgg.from_arg_list,
        "BIT_XOR": exp.BitwiseXorAgg.from_arg_list,
        "BIT_COUNT": exp.BitwiseCount.from_arg_list,
        "CURDATE": exp.CurrentDate.from_arg_list,
        "DATE_ADD": _build_dateadd,
        "DATEADD": _build_dateadd,
        "MAKE_TIMESTAMP": exp.TimestampFromParts.from_arg_list,
        "TIMESTAMPADD": _build_dateadd,
        "TIMESTAMPDIFF": build_date_delta(exp.TimestampDiff),
        "TRY_ADD": exp.SafeAdd.from_arg_list,
        "TRY_DIVIDE": exp.SafeDivide.from_arg_list,
        "TRY_MULTIPLY": exp.SafeMultiply.from_arg_list,
        "TRY_SUBTRACT": exp.SafeSubtract.from_arg_list,
        "DATEDIFF": _build_datediff,
        "DATE_DIFF": _build_datediff,
        "JSON_OBJECT_KEYS": exp.JSONKeys.from_arg_list,
        "LISTAGG": exp.GroupConcat.from_arg_list,
        "TIMESTAMP_LTZ": build_as_cast("TIMESTAMP_LTZ"),
        "TIMESTAMP_NTZ": build_as_cast("TIMESTAMP_NTZ"),
        "TRY_ELEMENT_AT": lambda args: exp.Bracket(
            this=seq_get(args, 0),
            expressions=ensure_list(seq_get(args, 1)),
            offset=1,
            safe=True,
        ),
        "LIKE": build_like(exp.Like),
        "ILIKE": build_like(exp.ILike),
    }

    PLACEHOLDER_PARSERS = {
        **Spark2Parser.PLACEHOLDER_PARSERS,
        TokenType.L_BRACE: lambda self: self._parse_query_parameter(),
    }

    def _parse_query_parameter(self) -> exp.Expr | None:
        this = self._parse_id_var()
        self._match(TokenType.R_BRACE)
        return self.expression(exp.Placeholder(this=this, widget=True))

    FUNCTION_PARSERS = {
        **Spark2Parser.FUNCTION_PARSERS,
        "SUBSTR": lambda self: self._parse_substring(),
    }

    STATEMENT_PARSERS = {
        **Spark2Parser.STATEMENT_PARSERS,
    }

    def _parse_generated_as_identity(
        self,
    ) -> (
        exp.GeneratedAsIdentityColumnConstraint
        | exp.ComputedColumnConstraint
        | exp.GeneratedAsRowColumnConstraint
    ):
        this = super()._parse_generated_as_identity()
        if this.expression:
            return self.expression(exp.ComputedColumnConstraint(this=this.expression))
        return this

    def _parse_pivot_aggregation(self) -> exp.Expr | None:
        # Spark 3+ and Databricks support non aggregate functions in PIVOT too, e.g
        # PIVOT (..., 'foo' AS bar FOR col_to_pivot IN (...))
        aggregate_expr = self._parse_function() or self._parse_disjunction()
        return self._parse_alias(aggregate_expr)



from __future__ import annotations

from sqlglot import exp
from sqlglot import generator
from sqlglot.dialects.dialect import (
    array_append_sql,
    rename_func,
    unit_to_var,
    timestampdiff_sql,
    date_delta_to_binary_interval_op,
    groupconcat_sql,
)
from sqlglot.generators.spark2 import Spark2Generator, temporary_storage_provider
from sqlglot.helper import seq_get
from sqlglot.transforms import (
    ctas_with_tmp_tables_to_create_tmp_view,
    remove_unique_constraints,
    preprocess,
    move_partitioned_by_to_schema_columns,
)


def _normalize_partition(e: exp.Expr) -> exp.Expr:
    """Normalize the expressions in PARTITION BY (<expression>, <expression>, ...)"""
    if isinstance(e, str):
        return exp.to_identifier(e)
    if isinstance(e, exp.Literal):
        return exp.to_identifier(e.name)
    return e


def _dateadd_sql(self: SparkGenerator, expression: exp.TsOrDsAdd | exp.TimestampAdd) -> str:
    if not expression.unit or (
        isinstance(expression, exp.TsOrDsAdd) and expression.text("unit").upper() == "DAY"
    ):
        # Coming from Hive/Spark2 DATE_ADD or roundtripping the 2-arg version of Spark3/DB
        return self.func("DATE_ADD", expression.this, expression.expression)

    this = self.func(
        "DATE_ADD",
        unit_to_var(expression),
        expression.expression,
        expression.this,
    )

    if isinstance(expression, exp.TsOrDsAdd):
        # The 3 arg version of DATE_ADD produces a timestamp in Spark3/DB but possibly not
        # in other dialects
        return_type = expression.return_type
        if not return_type.is_type(exp.DType.TIMESTAMP, exp.DType.DATETIME):
            this = f"CAST({this} AS {return_type})"

    return this


def _groupconcat_sql(self: SparkGenerator, expression: exp.GroupConcat) -> str:
    if self.dialect.version < (4,):
        expr = exp.ArrayToString(
            this=exp.ArrayAgg(this=expression.this),
            expression=expression.args.get("separator") or exp.Literal.string(""),
        )
        return self.sql(expr)

    return groupconcat_sql(self, expression)


class SparkGenerator(Spark2Generator):
    SUPPORTS_TO_NUMBER = True
    PAD_FILL_PATTERN_IS_REQUIRED = False
    SUPPORTS_CONVERT_TIMEZONE = True
    SUPPORTS_MEDIAN = True
    SUPPORTS_UNIX_SECONDS = True
    SUPPORTS_DECODE_CASE = True
    SET_ASSIGNMENT_REQUIRES_VARIABLE_KEYWORD = True

    TYPE_MAPPING = {
        **Spark2Generator.TYPE_MAPPING,
        exp.DType.MONEY: "DECIMAL",
        exp.DType.SMALLMONEY: "DECIMAL",
        exp.DType.UUID: "STRING",
        exp.DType.TIMESTAMPLTZ: "TIMESTAMP_LTZ",
        exp.DType.TIMESTAMPNTZ: "TIMESTAMP_NTZ",
    }

    TYPE_PARAM_SETTINGS = {
        **Spark2Generator.TYPE_PARAM_SETTINGS,
        exp.DType.MONEY: ((15, 4), ()),
        exp.DType.SMALLMONEY: ((6, 4), ()),
    }

    TRANSFORMS = {
        k: v
        for k, v in {
            **Spark2Generator.TRANSFORMS,
            exp.ArrayConstructCompact: lambda self, e: self.func(
                "ARRAY_COMPACT", self.func("ARRAY", *e.expressions)
            ),
            exp.ArrayInsert: lambda self, e: self.func(
                "ARRAY_INSERT", e.this, e.args.get("position"), e.expression
            ),
            exp.ArrayAppend: array_append_sql("ARRAY_APPEND"),
            exp.ArrayPrepend: array_append_sql("ARRAY_PREPEND"),
            exp.BitwiseAndAgg: rename_func("BIT_AND"),
            exp.BitwiseOrAgg: rename_func("BIT_OR"),
            exp.BitwiseXorAgg: rename_func("BIT_XOR"),
            exp.BitwiseCount: rename_func("BIT_COUNT"),
            exp.Create: preprocess(
                [
                    remove_unique_constraints,
                    lambda e: ctas_with_tmp_tables_to_create_tmp_view(
                        e, temporary_storage_provider
                    ),
                    move_partitioned_by_to_schema_columns,
                ]
            ),
            exp.CurrentVersion: rename_func("VERSION"),
            exp.DateFromUnixDate: rename_func("DATE_FROM_UNIX_DATE"),
            exp.DatetimeAdd: date_delta_to_binary_interval_op(cast=False),
            exp.DatetimeSub: date_delta_to_binary_interval_op(cast=False),
            exp.GroupConcat: _groupconcat_sql,
            exp.EndsWith: rename_func("ENDSWITH"),
            exp.JSONKeys: rename_func("JSON_OBJECT_KEYS"),
            exp.PartitionedByProperty: lambda self, e: (
                f"PARTITIONED BY {self.wrap(self.expressions(sqls=[_normalize_partition(e) for e in e.this.expressions], skip_first=True))}"
            ),
            exp.SafeAdd: rename_func("TRY_ADD"),
            exp.SafeDivide: rename_func("TRY_DIVIDE"),
            exp.SafeMultiply: rename_func("TRY_MULTIPLY"),
            exp.SafeSubtract: rename_func("TRY_SUBTRACT"),
            exp.StartsWith: rename_func("STARTSWITH"),
            exp.TimeAdd: date_delta_to_binary_interval_op(cast=False),
            exp.TimeSub: date_delta_to_binary_interval_op(cast=False),
            exp.TsOrDsAdd: _dateadd_sql,
            exp.TimestampAdd: _dateadd_sql,
            exp.TimestampFromParts: rename_func("MAKE_TIMESTAMP"),
            exp.TimestampSub: date_delta_to_binary_interval_op(cast=False),
            exp.DatetimeDiff: timestampdiff_sql,
            exp.TimestampDiff: timestampdiff_sql,
            exp.TryCast: lambda self, e: (
                self.trycast_sql(e) if e.args.get("safe") else self.cast_sql(e)
            ),
            exp.AnyValue: None,
            exp.DateDiff: None,
            exp.With: None,
        }.items()
        if v is not None
    }

    def ignorenulls_sql(self, expression: exp.IgnoreNulls) -> str:
        return generator.Generator.ignorenulls_sql(self, expression)

    def bracket_sql(self, expression: exp.Bracket) -> str:
        if expression.args.get("safe"):
            key = seq_get(self.bracket_offset_expressions(expression, index_offset=1), 0)
            return self.func("TRY_ELEMENT_AT", expression.this, key)

        return super().bracket_sql(expression)

    def computedcolumnconstraint_sql(self, expression: exp.ComputedColumnConstraint) -> str:
        return f"GENERATED ALWAYS AS ({self.sql(expression, 'this')})"

    def anyvalue_sql(self, expression: exp.AnyValue) -> str:
        return self.function_fallback_sql(expression)

    def datediff_sql(self, expression: exp.DateDiff) -> str:
        end = self.sql(expression, "this")
        start = self.sql(expression, "expression")

        if expression.unit:
            return self.func("DATEDIFF", unit_to_var(expression), start, end)

        return self.func("DATEDIFF", end, start)

    def placeholder_sql(self, expression: exp.Placeholder) -> str:
        if not expression.args.get("widget"):
            return super().placeholder_sql(expression)

        return f"{{{expression.name}}}"

    def readparquet_sql(self, expression: exp.ReadParquet) -> str:
        if len(expression.expressions) != 1:
            self.unsupported("READ_PARQUET with multiple arguments is not supported")
            return ""

        parquet_file = expression.expressions[0]
        return f"parquet.`{parquet_file.name}`"

    def ifblock_sql(self, expression: exp.IfBlock) -> str:
        condition = expression.this
        true_block = expression.args.get("true")

        condition_expr = None
        if isinstance(condition, exp.Not):
            inner = condition.this
            if isinstance(inner, exp.Is) and isinstance(inner.expression, exp.Null):
                condition_expr = inner.this

        if isinstance(condition_expr, exp.ObjectId):
            object_type = condition_expr.expression
            if (
                (object_type is None or object_type.name.upper() == "U")
                and isinstance(true_block, exp.Block)
                and isinstance(drop := true_block.expressions[0], exp.Drop)
            ):
                drop.set("exists", True)
                return self.sql(drop)

        return super().ifblock_sql(expression)



from __future__ import annotations

from sqlglot import exp
from sqlglot.typing import EXPRESSION_METADATA

EXPRESSION_METADATA = {
    **EXPRESSION_METADATA,
    **{
        expr_type: {"returns": exp.DType.UBIGINT}
        for expr_type in {
            exp.CountIf,
        }
    },
    exp.MD5Digest: {
        "annotator": lambda self, e: self._set_type(
            e, exp.DataType.build("FixedString(16)", dialect="clickhouse")
        )
    },
    exp.Corr: {
        "annotator": lambda self, e: self._set_type(
            e, exp.DataType.build("Float64", dialect="clickhouse")
        )
    },
}



from __future__ import annotations


from sqlglot import exp, tokens
from sqlglot.dialects.dialect import (
    Dialect,
    NormalizationStrategy,
)
from sqlglot.generators.clickhouse import ClickHouseGenerator
from sqlglot.parsers.clickhouse import ClickHouseParser
from sqlglot.tokens import TokenType
from sqlglot.typing.clickhouse import EXPRESSION_METADATA


class ClickHouse(Dialect):
    INDEX_OFFSET = 1
    NORMALIZE_FUNCTIONS: bool | str = False
    NULL_ORDERING = "nulls_are_last"
    SUPPORTS_USER_DEFINED_TYPES = False
    LOG_BASE_FIRST: bool | None = None
    FORCE_EARLY_ALIAS_REF_EXPANSION = True
    PRESERVE_ORIGINAL_NAMES = True
    NUMBERS_CAN_BE_UNDERSCORE_SEPARATED = True
    IDENTIFIERS_CAN_START_WITH_DIGIT = True
    HEX_STRING_IS_INTEGER_TYPE = True

    # https://github.com/ClickHouse/ClickHouse/issues/33935#issue-1112165779
    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_SENSITIVE

    EXPRESSION_METADATA = EXPRESSION_METADATA.copy()

    UNESCAPED_SEQUENCES = {
        "\\0": "\0",
    }

    CREATABLE_KIND_MAPPING = {"DATABASE": "SCHEMA"}

    SET_OP_DISTINCT_BY_DEFAULT: dict[type[exp.Expr], bool | None] = {
        exp.Except: False,
        exp.Intersect: False,
        exp.Union: None,
    }

    def generate_values_aliases(self, expression: exp.Values) -> list[exp.Identifier]:
        # Clickhouse allows VALUES to have an embedded structure e.g:
        # VALUES('person String, place String', ('Noah', 'Paris'), ...)
        # In this case, we don't want to qualify the columns
        values = expression.expressions[0].expressions

        structure = (
            values[0]
            if (len(values) > 1 and values[0].is_string and isinstance(values[1], exp.Tuple))
            else None
        )
        if structure:
            # Split each column definition into the column name e.g:
            # 'person String, place String' -> ['person', 'place']
            structure_coldefs = [coldef.strip() for coldef in structure.name.split(",")]
            column_aliases = [
                exp.to_identifier(coldef.split(" ")[0]) for coldef in structure_coldefs
            ]
        else:
            # Default column aliases in CH are "c1", "c2", etc.
            column_aliases = [
                exp.to_identifier(f"c{i + 1}") for i in range(len(values[0].expressions))
            ]

        return column_aliases

    class Tokenizer(tokens.Tokenizer):
        COMMENTS = ["--", "#", "#!", ("/*", "*/")]
        IDENTIFIERS = ['"', "`"]
        IDENTIFIER_ESCAPES = ["\\"]
        STRING_ESCAPES = ["'", "\\"]
        BIT_STRINGS = [("0b", "")]
        HEX_STRINGS = [("0x", ""), ("0X", "")]
        HEREDOC_STRINGS = ["$"]

        KEYWORDS = {
            **tokens.Tokenizer.KEYWORDS,
            ".:": TokenType.DOTCOLON,
            ".^": TokenType.DOTCARET,
            "ATTACH": TokenType.COMMAND,
            "DATE32": TokenType.DATE32,
            "DETACH": TokenType.DETACH,
            "DATETIME64": TokenType.DATETIME64,
            "DICTIONARY": TokenType.DICTIONARY,
            "DYNAMIC": TokenType.DYNAMIC,
            "ENUM8": TokenType.ENUM8,
            "ENUM16": TokenType.ENUM16,
            "EXCHANGE": TokenType.COMMAND,
            "FINAL": TokenType.FINAL,
            "FIXEDSTRING": TokenType.FIXEDSTRING,
            "FLOAT32": TokenType.FLOAT,
            "FLOAT64": TokenType.DOUBLE,
            "GLOBAL": TokenType.GLOBAL,
            "LOWCARDINALITY": TokenType.LOWCARDINALITY,
            "MAP": TokenType.MAP,
            "NESTED": TokenType.NESTED,
            "NOTHING": TokenType.NOTHING,
            "SAMPLE": TokenType.TABLE_SAMPLE,
            "TUPLE": TokenType.STRUCT,
            "UINT16": TokenType.USMALLINT,
            "UINT32": TokenType.UINT,
            "UINT64": TokenType.UBIGINT,
            "UINT8": TokenType.UTINYINT,
            "IPV4": TokenType.IPV4,
            "IPV6": TokenType.IPV6,
            "POINT": TokenType.POINT,
            "PROJECTION": TokenType.PROJECTION,
            "RING": TokenType.RING,
            "LINESTRING": TokenType.LINESTRING,
            "MULTILINESTRING": TokenType.MULTILINESTRING,
            "POLYGON": TokenType.POLYGON,
            "MULTIPOLYGON": TokenType.MULTIPOLYGON,
            "AGGREGATEFUNCTION": TokenType.AGGREGATEFUNCTION,
            "SIMPLEAGGREGATEFUNCTION": TokenType.SIMPLEAGGREGATEFUNCTION,
            "SYSTEM": TokenType.COMMAND,
            "PREWHERE": TokenType.PREWHERE,
        }

        KEYWORDS.pop("/*+")

        SINGLE_TOKENS = {
            **tokens.Tokenizer.SINGLE_TOKENS,
            "$": TokenType.HEREDOC_STRING,
        }

    Parser = ClickHouseParser

    Generator = ClickHouseGenerator



from __future__ import annotations

from sqlglot import exp
from sqlglot.typing.spark import EXPRESSION_METADATA

EXPRESSION_METADATA = {
    **EXPRESSION_METADATA,
    **{
        exp_type: {"returns": exp.DType.DOUBLE}
        for exp_type in {
            exp.RegrAvgx,
            exp.RegrAvgy,
            exp.RegrIntercept,
            exp.RegrR2,
            exp.RegrSlope,
            exp.RegrSxx,
            exp.RegrSxy,
            exp.RegrSyy,
            exp.Rint,
        }
    },
    **{
        exp_type: {"returns": exp.DType.INT}
        for exp_type in {
            exp.RegexpCount,
            exp.RegexpInstr,
        }
    },
    **{
        exp_type: {"returns": exp.DType.VARCHAR}
        for exp_type in {
            exp.RegexpSubstr,
            exp.Secret,
        }
    },
    exp.RegrCount: {"returns": exp.DType.BIGINT},
    exp.Search: {"returns": exp.DType.BOOLEAN},
    exp.RegexpExtractAll: {
        "annotator": lambda self, e: self._set_type(
            e, exp.DataType.from_str("ARRAY<STRING>", dialect="databricks")
        )
    },
}
